/*
 * Animation plugin for beryl
 * 
 * Copyright (C) 2006 Erkin Bahceci
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <X11/Xatom.h>

#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <compiz.h>

#define FAKE_ICON_SIZE 4

#define WIN_X(w) ((w)->attrib.x - (w)->output.left)
#define WIN_Y(w) ((w)->attrib.y - (w)->output.top)
#define WIN_W(w) ((w)->width + (w)->output.left + (w)->output.right)
#define WIN_H(w) ((w)->height + (w)->output.top + (w)->output.bottom)

#define BORDER_X(w) ((w)->attrib.x - (w)->input.left)
#define BORDER_Y(w) ((w)->attrib.y - (w)->input.top)
#define BORDER_W(w) ((w)->width + (w)->input.left + (w)->input.right)
#define BORDER_H(w) ((w)->height + (w)->input.top + (w)->input.bottom)

typedef struct _xy_pair
{
    float x, y;
} Point, Vector;

typedef struct _Object
{
    Point gridPosition;           // position on window in [0,1] range
    Point position;               // position on screen

    // Texture x, y coordinates will be offset by given amounts
    // for quads that fall after and before this object in x and y directions.
    // Currently only y offset can be used.
    Point offsetTexCoordForQuadBefore;
    Point offsetTexCoordForQuadAfter;
} Object;

typedef struct _WaveParam
{
    float halfWidth;
    float amp;
    float pos;
} WaveParam;

typedef enum
{
    WindowEventNone = 0,
    WindowEventMinimize,
    WindowEventUnminimize,
    WindowEventClose,
    WindowEventCreate,
    WindowEventFocus,
    WindowEventShade,
    WindowEventUnshade
} WindowEvent;

typedef struct _Model
{
    Object *objects;
    int numObjects;
    int gridWidth;
    int gridHeight;
    float remainderSteps;
    Vector scale;
    Point scaleOrigin;
    Bool transformed;
    Point topLeft;
    Point bottomRight;

    int magicLampWaveCount;
    WaveParam *magicLampWaves;
    WindowEvent forWindowEvent;
    float topHeight;
    float bottomHeight;
} Model;

#define ANIM_MAGIC_LAMP1_GRID_RES_DEFAULT  100
#define ANIM_MAGIC_LAMP1_GRID_RES_MIN      4
#define ANIM_MAGIC_LAMP1_GRID_RES_MAX      200

#define ANIM_MAGIC_LAMP1_MAX_WAVES_DEFAULT  5
#define ANIM_MAGIC_LAMP1_MAX_WAVES_MIN      3
#define ANIM_MAGIC_LAMP1_MAX_WAVES_MAX      20

#define ANIM_MAGIC_LAMP1_WAVE_AMP_MIN_DEFAULT  200
#define ANIM_MAGIC_LAMP1_WAVE_AMP_MIN_MIN      200
#define ANIM_MAGIC_LAMP1_WAVE_AMP_MIN_MAX      2000
#define ANIM_MAGIC_LAMP1_WAVE_AMP_MIN_PRECISION 5

#define ANIM_MAGIC_LAMP1_WAVE_AMP_MAX_DEFAULT  300
#define ANIM_MAGIC_LAMP1_WAVE_AMP_MAX_MIN      200
#define ANIM_MAGIC_LAMP1_WAVE_AMP_MAX_MAX      2000
#define ANIM_MAGIC_LAMP1_WAVE_AMP_MAX_PRECISION 5

#define ANIM_MAGIC_LAMP2_GRID_RES_DEFAULT  100
#define ANIM_MAGIC_LAMP2_GRID_RES_MIN      2
#define ANIM_MAGIC_LAMP2_GRID_RES_MAX      200

#define ANIM_MAGIC_LAMP2_MAX_WAVES_DEFAULT  3
#define ANIM_MAGIC_LAMP2_MAX_WAVES_MIN      3
#define ANIM_MAGIC_LAMP2_MAX_WAVES_MAX      20

#define ANIM_MAGIC_LAMP2_WAVE_AMP_MIN_DEFAULT  200
#define ANIM_MAGIC_LAMP2_WAVE_AMP_MIN_MIN      200
#define ANIM_MAGIC_LAMP2_WAVE_AMP_MIN_MAX      2000
#define ANIM_MAGIC_LAMP2_WAVE_AMP_MIN_PRECISION 5

#define ANIM_MAGIC_LAMP2_WAVE_AMP_MAX_DEFAULT  300
#define ANIM_MAGIC_LAMP2_WAVE_AMP_MAX_MIN      200
#define ANIM_MAGIC_LAMP2_WAVE_AMP_MAX_MAX      2000
#define ANIM_MAGIC_LAMP2_WAVE_AMP_MAX_PRECISION 5

#define ANIM_MINIMIZE_DURATION_DEFAULT   1.0
#define ANIM_MINIMIZE_DURATION_MIN       0.1
#define ANIM_MINIMIZE_DURATION_MAX       10
#define ANIM_MINIMIZE_DURATION_PRECISION 0.05

#define ANIM_UNMINIMIZE_DURATION_DEFAULT   1.0
#define ANIM_UNMINIMIZE_DURATION_MIN       0.1
#define ANIM_UNMINIMIZE_DURATION_MAX       10
#define ANIM_UNMINIMIZE_DURATION_PRECISION 0.05

#define ANIM_CREATE1_DURATION_DEFAULT   0.7
#define ANIM_CREATE1_DURATION_MIN       0.1
#define ANIM_CREATE1_DURATION_MAX       10
#define ANIM_CREATE1_DURATION_PRECISION 0.05

#define ANIM_CLOSE1_DURATION_DEFAULT   0.7
#define ANIM_CLOSE1_DURATION_MIN       0.1
#define ANIM_CLOSE1_DURATION_MAX       10
#define ANIM_CLOSE1_DURATION_PRECISION 0.05

#define ANIM_CREATE2_DURATION_DEFAULT   0.7
#define ANIM_CREATE2_DURATION_MIN       0.1
#define ANIM_CREATE2_DURATION_MAX       10
#define ANIM_CREATE2_DURATION_PRECISION 0.05

#define ANIM_CLOSE2_DURATION_DEFAULT   0.7
#define ANIM_CLOSE2_DURATION_MIN       0.1
#define ANIM_CLOSE2_DURATION_MAX       10
#define ANIM_CLOSE2_DURATION_PRECISION 0.05

#define ANIM_FOCUS_DURATION_DEFAULT   0.6
#define ANIM_FOCUS_DURATION_MIN       0.1
#define ANIM_FOCUS_DURATION_MAX       10
#define ANIM_FOCUS_DURATION_PRECISION 0.05

#define ANIM_SHADE_DURATION_DEFAULT   1
#define ANIM_SHADE_DURATION_MIN       0.1
#define ANIM_SHADE_DURATION_MAX       10
#define ANIM_SHADE_DURATION_PRECISION 0.05

#define ANIM_UNSHADE_DURATION_DEFAULT   1
#define ANIM_UNSHADE_DURATION_MIN       0.1
#define ANIM_UNSHADE_DURATION_MAX       10
#define ANIM_UNSHADE_DURATION_PRECISION 0.05

#define ANIM_FOCUS_WAVE_WIDTH_DEFAULT   0.7
#define ANIM_FOCUS_WAVE_WIDTH_MIN       0.02
#define ANIM_FOCUS_WAVE_WIDTH_MAX       3
#define ANIM_FOCUS_WAVE_WIDTH_PRECISION 0.02

#define ANIM_FOCUS_WAVE_AMP_DEFAULT   0.03
#define ANIM_FOCUS_WAVE_AMP_MIN       0
#define ANIM_FOCUS_WAVE_AMP_MAX       1
#define ANIM_FOCUS_WAVE_AMP_PRECISION 0.01

#define ANIM_CURVED_FOLD_AMP_DEFAULT   0.15
#define ANIM_CURVED_FOLD_AMP_MIN      -0.5
#define ANIM_CURVED_FOLD_AMP_MAX       0.5
#define ANIM_CURVED_FOLD_AMP_PRECISION 0.01

#define ANIM_HORIZONTAL_FOLDS_NUM_FOLDS_DEFAULT  5
#define ANIM_HORIZONTAL_FOLDS_NUM_FOLDS_MIN      1
#define ANIM_HORIZONTAL_FOLDS_NUM_FOLDS_MAX      50

#define ANIM_HORIZONTAL_FOLDS_AMP_DEFAULT   0.07
#define ANIM_HORIZONTAL_FOLDS_AMP_MIN      -0.5
#define ANIM_HORIZONTAL_FOLDS_AMP_MAX       0.5
#define ANIM_HORIZONTAL_FOLDS_AMP_PRECISION 0.01

#define ANIM_TIME_STEP_DEFAULT   25
#define ANIM_TIME_STEP_MIN       10
#define ANIM_TIME_STEP_MAX       400

#define ANIM_ROLLUP_FIXED_INTERIOR_DEFAULT  FALSE


typedef enum
{
    PostprocessDisablingNone = 0,
    PostprocessDisablingWindow,
    PostprocessDisablingScreen
} PostprocessDisabling;

static char *ppDisablingName[] = {
    N_("No disabling"),
    N_("Animated window"),
    N_("All windows")
};

#define ANIM_DISABLE_PP_FX_DEFAULT PostprocessDisablingScreen

typedef enum
{
    AnimEffectNone = 0,
    AnimEffectFade,
    AnimEffectMagicLamp1,
    AnimEffectMagicLamp2,
    AnimEffectZoom,
    AnimEffectSidekick,
    AnimEffectDream,
    AnimEffectWave,
    AnimEffectCurvedFold,
    AnimEffectHorizontalFolds,
    AnimEffectRollUp,
    AnimEffectNum
} AnimEffect;

static char *allEffectName[] = {
    N_("None"),
    N_("Fade"),
    N_("Magic Lamp 1"),
    N_("Magic Lamp 2"),
    N_("Zoom"),
    N_("Sidekick"),
    N_("Dream"),
    N_("Wave"),
    N_("Curved Fold"),
    N_("Horizontal Folds"),
    N_("Roll Up")
};

#define ANIM_MINIMIZE_DEFAULT   AnimEffectMagicLamp2
#define ANIM_UNMINIMIZE_DEFAULT AnimEffectMagicLamp2
#define ANIM_CLOSE1_DEFAULT     AnimEffectZoom
#define ANIM_CREATE1_DEFAULT    AnimEffectZoom
#define ANIM_CLOSE2_DEFAULT     AnimEffectDream
#define ANIM_CREATE2_DEFAULT    AnimEffectDream
#define ANIM_FOCUS_DEFAULT      AnimEffectNone
#define ANIM_SHADE_DEFAULT      AnimEffectCurvedFold
#define ANIM_UNSHADE_DEFAULT    AnimEffectCurvedFold

static char *minimizeEffectName[] = {
    N_("None"),
    N_("Fade"),
    N_("Magic Lamp 1"),
    N_("Magic Lamp 2"),
    N_("Zoom"),
    N_("Sidekick"),
    N_("Dream"),
    N_("Wave"),
    N_("Curved Fold"),
    N_("Horizontal Folds")
};

static AnimEffect minimizeEffectType[] = {
    AnimEffectNone,
    AnimEffectFade,
    AnimEffectMagicLamp1,
    AnimEffectMagicLamp2,
    AnimEffectZoom,
    AnimEffectSidekick,
    AnimEffectDream,
    AnimEffectWave,
    AnimEffectCurvedFold,
    AnimEffectHorizontalFolds
};

#define NUM_MINIMIZE_EFFECT (sizeof (minimizeEffectType) / sizeof (minimizeEffectType[0]))

static char *closeEffectName[] = {
    N_("None"),
    N_("Fade"),
    N_("Zoom"),
    N_("Sidekick"),
    N_("Dream"),
    N_("Wave"),
    N_("Curved Fold"),
    N_("Horizontal Folds")
};

static AnimEffect closeEffectType[] = {
    AnimEffectNone,
    AnimEffectFade,
    AnimEffectZoom,
    AnimEffectSidekick,
    AnimEffectDream,
    AnimEffectWave,
    AnimEffectCurvedFold,
    AnimEffectHorizontalFolds
};

#define NUM_CLOSE_EFFECT (sizeof (closeEffectType) / sizeof (closeEffectType[0]))

static char *focusEffectName[] = {
    N_("None"),
    N_("Wave")
};

static AnimEffect focusEffectType[] = {
    AnimEffectNone,
    AnimEffectWave
};

#define NUM_FOCUS_EFFECT (sizeof (focusEffectType) / sizeof (focusEffectType[0]))

static char *shadeEffectName[] = {
    N_("None"),
    N_("Curved Fold"),
    N_("Horizontal Folds"),
    N_("Roll Up")
};

static AnimEffect shadeEffectType[] = {
    AnimEffectNone,
    AnimEffectCurvedFold,
    AnimEffectHorizontalFolds,
    AnimEffectRollUp
};

#define NUM_SHADE_EFFECT (sizeof (shadeEffectType) / sizeof (shadeEffectType[0]))

static char *minimizeDefaultWinType[] = {
    N_("Utility"),
    N_("Dialog"),
    N_("Dnd"),
    N_("Unknown"),
    N_("ModalDialog"),
    N_("Normal")
};

#define N_MINIMIZE_DEFAULT_WIN_TYPE (sizeof (minimizeDefaultWinType) / sizeof (minimizeDefaultWinType[0]))

static char *close1DefaultWinType[] = {
    N_("Normal"),
    N_("Dialog"),
    N_("ModalDialog")
};

#define N_CLOSE1_DEFAULT_WIN_TYPE (sizeof (close1DefaultWinType) / sizeof (close1DefaultWinType[0]))

static char *close2DefaultWinType[] = {
    N_("Unknown"),
    N_("Menu"),
    N_("PopupMenu"),
    N_("DropdownMenu"),
    N_("Tooltip"),
    N_("Dnd"),
    N_("Combo")
};

#define N_CLOSE2_DEFAULT_WIN_TYPE (sizeof (close2DefaultWinType) / sizeof (close2DefaultWinType[0]))

static char *focusDefaultWinType[] = {
    N_("Normal"),
    N_("Dialog"),
    N_("ModalDialog"),
    N_("Dnd")
};

#define N_FOCUS_DEFAULT_WIN_TYPE (sizeof (focusDefaultWinType) / sizeof (focusDefaultWinType[0]))

static char *shadeDefaultWinType[] = {
    N_("Normal"),
    N_("Dialog"),
    N_("ModalDialog"),
    N_("Dnd")
};

#define N_SHADE_DEFAULT_WIN_TYPE (sizeof (shadeDefaultWinType) / sizeof (shadeDefaultWinType[0]))

#define ANIM_ZOOM_FROM_CENTER_DEFAULT FALSE

static int displayPrivateIndex;

typedef struct _AnimDisplay
{
    int screenPrivateIndex;
    Atom wmHintsAtom;
    Atom winIconGeometryAtom;
    HandleEventProc handleEvent;
} AnimDisplay;

#define ANIM_SCREEN_OPTION_MINIMIZE_EFFECT            0
#define ANIM_SCREEN_OPTION_MINIMIZE_WINDOW_TYPE       1
#define ANIM_SCREEN_OPTION_MINIMIZE_DURATION          2
#define ANIM_SCREEN_OPTION_UNMINIMIZE_EFFECT          3
#define ANIM_SCREEN_OPTION_UNMINIMIZE_WINDOW_TYPE     4
#define ANIM_SCREEN_OPTION_UNMINIMIZE_DURATION        5
#define ANIM_SCREEN_OPTION_CLOSE1_EFFECT              6
#define ANIM_SCREEN_OPTION_CLOSE1_WINDOW_TYPE         7
#define ANIM_SCREEN_OPTION_CLOSE1_DURATION            8
#define ANIM_SCREEN_OPTION_CREATE1_EFFECT             9
#define ANIM_SCREEN_OPTION_CREATE1_WINDOW_TYPE       10
#define ANIM_SCREEN_OPTION_CREATE1_DURATION          11
#define ANIM_SCREEN_OPTION_CLOSE2_EFFECT             12
#define ANIM_SCREEN_OPTION_CLOSE2_WINDOW_TYPE        13
#define ANIM_SCREEN_OPTION_CLOSE2_DURATION           14
#define ANIM_SCREEN_OPTION_CREATE2_EFFECT            15
#define ANIM_SCREEN_OPTION_CREATE2_WINDOW_TYPE       16
#define ANIM_SCREEN_OPTION_CREATE2_DURATION          17
#define ANIM_SCREEN_OPTION_FOCUS_EFFECT              18
#define ANIM_SCREEN_OPTION_FOCUS_WINDOW_TYPE         19
#define ANIM_SCREEN_OPTION_FOCUS_DURATION            20
#define ANIM_SCREEN_OPTION_SHADE_EFFECT              21
#define ANIM_SCREEN_OPTION_SHADE_WINDOW_TYPE         22
#define ANIM_SCREEN_OPTION_SHADE_DURATION            23
#define ANIM_SCREEN_OPTION_UNSHADE_EFFECT            24
#define ANIM_SCREEN_OPTION_UNSHADE_WINDOW_TYPE       25
#define ANIM_SCREEN_OPTION_UNSHADE_DURATION          26
#define ANIM_SCREEN_OPTION_MAGIC_LAMP1_GRID_RES      27
#define ANIM_SCREEN_OPTION_MAGIC_LAMP1_MAX_WAVES     28
#define ANIM_SCREEN_OPTION_MAGIC_LAMP1_WAVE_AMP_MIN  29
#define ANIM_SCREEN_OPTION_MAGIC_LAMP1_WAVE_AMP_MAX  30
#define ANIM_SCREEN_OPTION_MAGIC_LAMP2_GRID_RES      31
#define ANIM_SCREEN_OPTION_MAGIC_LAMP2_MAX_WAVES     32
#define ANIM_SCREEN_OPTION_MAGIC_LAMP2_WAVE_AMP_MIN  33
#define ANIM_SCREEN_OPTION_MAGIC_LAMP2_WAVE_AMP_MAX  34
#define ANIM_SCREEN_OPTION_FOCUS_WAVE_WIDTH          35
#define ANIM_SCREEN_OPTION_FOCUS_WAVE_AMP            36
#define ANIM_SCREEN_OPTION_CURVED_FOLD_AMP           37
#define ANIM_SCREEN_OPTION_HORIZONTAL_FOLDS_AMP      38
#define ANIM_SCREEN_OPTION_HORIZONTAL_FOLDS_NUM_FOLDS 39
#define ANIM_SCREEN_OPTION_ZOOM_FROM_CENTER          40
#define ANIM_SCREEN_OPTION_ROLLUP_FIXED_INTERIOR     41
#define ANIM_SCREEN_OPTION_DISABLE_PP_FX             42
#define ANIM_SCREEN_OPTION_TIME_STEP                 43
#define ANIM_SCREEN_OPTION_NUM                       44

typedef struct _AnimScreen
{
    int windowPrivateIndex;

    PreparePaintScreenProc preparePaintScreen;
    DonePaintScreenProc donePaintScreen;
    PaintScreenProc paintScreen;
    PaintWindowProc paintWindow;
    DamageWindowRectProc damageWindowRect;
    AddWindowGeometryProc addWindowGeometry;
    DrawWindowGeometryProc drawWindowGeometry;

    WindowResizeNotifyProc windowResizeNotify;
    WindowMoveNotifyProc windowMoveNotify;
    WindowGrabNotifyProc windowGrabNotify;
    WindowUngrabNotifyProc windowUngrabNotify;
    SetWindowScaleProc setWindowScale;

    CompOption opt[ANIM_SCREEN_OPTION_NUM];

    PostprocessDisabling ppDisabling;

    Bool animInProgress;
    AnimEffect minimizeEffect;
    AnimEffect unminimizeEffect;
    AnimEffect create1Effect;
    AnimEffect create2Effect;
    AnimEffect close1Effect;
    AnimEffect close2Effect;
    AnimEffect focusEffect;
    AnimEffect shadeEffect;
    AnimEffect unshadeEffect;
    unsigned int minimizeWMask;
    unsigned int unminimizeWMask;
    unsigned int create1WMask;
    unsigned int create2WMask;
    unsigned int close1WMask;
    unsigned int close2WMask;
    unsigned int focusWMask;
    unsigned int shadeWMask;
    unsigned int unshadeWMask;
} AnimScreen;

typedef struct _AnimWindow
{
    Model *model;
    unsigned int state;
    unsigned int newState;

    XRectangle icon;
    XRectangle origWindowRect;

    float numZoomRotations;
    float storedOpacity;
    float timestep;               // to be used in updateWindowAttribFunc

    float animTotalTime;
    float animRemainingTime;
    int animOverrideProgressDir;  // 0: default dir, 1: forward, 2: backward

    Bool nowShaded;
    Bool grabbed;

    WindowEvent curWindowEvent;
    AnimEffect curAnimEffect;

    int unmapCnt;
    int destroyCnt;
} AnimWindow;

#define GET_ANIM_DISPLAY(d)                                       \
    ((AnimDisplay *) (d)->privates[displayPrivateIndex].ptr)

#define ANIM_DISPLAY(d)                       \
    AnimDisplay *ad = GET_ANIM_DISPLAY (d)

#define GET_ANIM_SCREEN(s, ad)                                   \
    ((AnimScreen *) (s)->privates[(ad)->screenPrivateIndex].ptr)

#define ANIM_SCREEN(s)                                                      \
    AnimScreen *as = GET_ANIM_SCREEN (s, GET_ANIM_DISPLAY (s->display))

#define GET_ANIM_WINDOW(w, as)                                   \
    ((AnimWindow *) (w)->privates[(as)->windowPrivateIndex].ptr)

#define ANIM_WINDOW(w)                                         \
    AnimWindow *aw = GET_ANIM_WINDOW  (w,                         \
            GET_ANIM_SCREEN  (w->screen,                 \
                GET_ANIM_DISPLAY (w->screen->display)))

#define NUM_OPTIONS(s) (sizeof ((s)->opt) / sizeof (CompOption))






static void
modelCalcBounds (Model * model)
{
    int i;

    model->topLeft.x = MAXSHORT;
    model->topLeft.y = MAXSHORT;
    model->bottomRight.x = MINSHORT;
    model->bottomRight.y = MINSHORT;

    for (i = 0; i < model->numObjects; i++)
    {
        if (model->objects[i].position.x < model->topLeft.x)
            model->topLeft.x = model->objects[i].position.x;
        else if (model->objects[i].position.x > model->bottomRight.x)
            model->bottomRight.x = model->objects[i].position.x;

        if (model->objects[i].position.y < model->topLeft.y)
            model->topLeft.y = model->objects[i].position.y;
        else if (model->objects[i].position.y > model->bottomRight.y)
            model->bottomRight.y = model->objects[i].position.y;
    }
}


// =====================  Effect: Magic Lamp  =========================

static void
fxMagicLamp1InitGrid (AnimScreen * as,
                      WindowEvent forWindowEvent,
                      int * gridWidth,
                      int * gridHeight)
{
    *gridWidth = 2;
    *gridHeight = as->opt[ANIM_SCREEN_OPTION_MAGIC_LAMP1_GRID_RES].value.i;
}

static void
fxMagicLamp2InitGrid (AnimScreen * as,
                      WindowEvent forWindowEvent,
                      int * gridWidth,
                      int * gridHeight)
{
    *gridWidth = 2;
    *gridHeight = as->opt[ANIM_SCREEN_OPTION_MAGIC_LAMP2_GRID_RES].value.i;
}

static void
fxMagicLampInit (CompScreen * s, CompWindow * w)
{
    ANIM_SCREEN (s);
    ANIM_WINDOW (w);

    Model *model = aw->model;

    int screenHeight = s->height;
    Bool minimizeToTop = (screenHeight - aw->icon.y > aw->icon.y);
    int maxWaves;
    float waveAmpMin, waveAmpMax;

    if (aw->curAnimEffect == AnimEffectMagicLamp1)
    {
        maxWaves = as->opt[ANIM_SCREEN_OPTION_MAGIC_LAMP1_MAX_WAVES].value.i;
        waveAmpMin =
            as->opt[ANIM_SCREEN_OPTION_MAGIC_LAMP1_WAVE_AMP_MIN].value.f;
        waveAmpMax =
            as->opt[ANIM_SCREEN_OPTION_MAGIC_LAMP1_WAVE_AMP_MAX].value.f;
    }
    else
    {
        maxWaves = as->opt[ANIM_SCREEN_OPTION_MAGIC_LAMP2_MAX_WAVES].value.i;
        waveAmpMin =
            as->opt[ANIM_SCREEN_OPTION_MAGIC_LAMP2_WAVE_AMP_MIN].value.f;
        waveAmpMax =
            as->opt[ANIM_SCREEN_OPTION_MAGIC_LAMP2_WAVE_AMP_MAX].value.f;
    }
    if (waveAmpMax < waveAmpMin)
        waveAmpMax = waveAmpMin;

    if (maxWaves > 0)
    {
        float distance;
        if (minimizeToTop)
            distance = WIN_Y (w) + WIN_H (w) - aw->icon.y;
        else
            distance = aw->icon.y - WIN_Y (w);

        // Initialize waves
        model->magicLampWaveCount =
            1 + (float)maxWaves * distance / screenHeight;

        if (!(model->magicLampWaves))
            model->magicLampWaves =
                malloc (model->magicLampWaveCount * sizeof (WaveParam));

        int ampDirection = ((float)rand () / RAND_MAX < 0.5 ? 1 : -1);
        int i;
        float minHalfWidth = 0.22f;
        float maxHalfWidth = 0.38f;

        for (i = 0; i < model->magicLampWaveCount; i++)
        {
            model->magicLampWaves[i].amp =
                ampDirection * (waveAmpMax - waveAmpMin) *
                rand () / RAND_MAX +
                ampDirection * waveAmpMin;
            model->magicLampWaves[i].halfWidth =
                (float)rand () / RAND_MAX * (maxHalfWidth - minHalfWidth) + minHalfWidth;

            // avoid offset at top and bottom part by added waves
            float availPos = 1 - 2 * model->magicLampWaves[i].halfWidth;
            float posInAvailSegment = 0;
            if (i > 0)
                posInAvailSegment = (availPos / model->magicLampWaveCount) * rand () / RAND_MAX;

            model->magicLampWaves[i].pos =
                (posInAvailSegment +
                 i * availPos / model->magicLampWaveCount +
                 model->magicLampWaves[i].halfWidth);

            // switch wave direction
            ampDirection *= -1;
        }
    }
    else
        model->magicLampWaveCount = 0;
}

#define sigmoid(fx) (1.0f/(1.0f+exp(-5.0f*2*((fx)-0.5))))
#define sigmoid2(fx, c) (1.0f/(1.0f+exp(-(c)*2*((fx)-0.5))))

static void
fxMagicLampModelStepObject (CompWindow * w,
                            Model * model,
                            Object * object,
                            float forwardProgress,
                            Bool minimizeToTop)
{
    ANIM_WINDOW (w);

    float iconCloseEndY;
    float winFarEndY;
    float winVisibleCloseEndY;

    if (minimizeToTop)
    {
        iconCloseEndY = aw->icon.y + aw->icon.height;
        winFarEndY = WIN_Y (w) + WIN_H (w);
        winVisibleCloseEndY = WIN_Y (w);
        if (winVisibleCloseEndY < iconCloseEndY)
            winVisibleCloseEndY = iconCloseEndY;
    }
    else
    {
        iconCloseEndY = aw->icon.y;
        winFarEndY = WIN_Y (w);
        winVisibleCloseEndY = WIN_Y (w) + WIN_H (w);
        if (winVisibleCloseEndY > iconCloseEndY)
            winVisibleCloseEndY = iconCloseEndY;
    }

    float preShapePhaseEnd = 0.17f;
    float stretchPhaseEnd =
        preShapePhaseEnd + (1 - preShapePhaseEnd) *
        (iconCloseEndY - winVisibleCloseEndY) / ((iconCloseEndY - winFarEndY) +
                                              (iconCloseEndY - winVisibleCloseEndY));
    if (stretchPhaseEnd < preShapePhaseEnd + 0.1)
        stretchPhaseEnd = preShapePhaseEnd + 0.1;

    float origx =
        WIN_X (w) +
        WIN_W (w) * object->gridPosition.x;
    float origy =
        WIN_Y (w) +
        WIN_H (w) * object->gridPosition.y;
    float iconShadowLeft = ((float)(w->output.left - w->input.left)) * aw->icon.width / w->width;
    float iconShadowRight = ((float)(w->output.right - w->input.right)) * aw->icon.width / w->width;
    float iconx =
        (aw->icon.x - iconShadowLeft) + (aw->icon.width + iconShadowLeft +
                                         iconShadowRight) *
        object->gridPosition.x;
    float icony = aw->icon.y + aw->icon.height * object->gridPosition.y;

    if (forwardProgress < preShapePhaseEnd)
    {
        float preShapeProgress = forwardProgress / preShapePhaseEnd;
        float fx = (iconCloseEndY - object->position.y) / (iconCloseEndY - winFarEndY);
        float fy = (sigmoid (fx) - sigmoid (0)) / (sigmoid (1) - sigmoid (0));
        int i;
        float targetx = fy * (origx - iconx) + iconx;
        for (i = 0; i < model->magicLampWaveCount; i++)
        {
            float cosfx =
                (fx -
                 model->magicLampWaves[i].pos) /
                model->magicLampWaves[i].halfWidth;
            if (cosfx < -1 || cosfx > 1)
                continue;
            targetx +=
                model->magicLampWaves[i].amp * (cos (cosfx * M_PI) + 1) / 2;
        }
        object->position.x =
            (1 - preShapeProgress) * origx + preShapeProgress * targetx;
        object->position.y = origy;
    }
    else
    {
        float stretchedPos;
        if (minimizeToTop)
            stretchedPos =
                object->gridPosition.y * origy +
                (1 - object->gridPosition.y) * icony;
        else
            stretchedPos =
                (1 - object->gridPosition.y) * origy +
                object->gridPosition.y * icony;

        if (forwardProgress < stretchPhaseEnd)
        {
            float stretchProgress =
                (forwardProgress - preShapePhaseEnd) /
                (stretchPhaseEnd - preShapePhaseEnd);

            object->position.y =
                (1 - stretchProgress) * origy +
                stretchProgress * stretchedPos;
        }
        else
        {
            float postStretchProgress =
                (forwardProgress - stretchPhaseEnd) / (1 - stretchPhaseEnd);

            object->position.y =
                (1 - postStretchProgress) *
                stretchedPos +
                postStretchProgress *
                (stretchedPos + (iconCloseEndY - winFarEndY));
        }
        float fx = (iconCloseEndY - object->position.y) / (iconCloseEndY - winFarEndY);
        float fy = (sigmoid (fx) - sigmoid (0)) / (sigmoid (1) - sigmoid (0));

        int i;
        object->position.x = fy * (origx - iconx) + iconx;
        for (i = 0; i < model->magicLampWaveCount; i++)
        {
            float cosfx =
                (fx -
                 model->magicLampWaves[i].pos) /
                model->magicLampWaves[i].halfWidth;
            if (cosfx < -1 || cosfx > 1)
                continue;
            object->position.x +=
                model->magicLampWaves[i].amp * (cos (cosfx * M_PI) + 1) / 2;
        }
    }
}

static void
fxMagicLampModelStep (CompScreen * s, CompWindow * w, float time)
{
    int i, j, steps;

    ANIM_SCREEN (s);
    ANIM_WINDOW (w);

    Model *model = aw->model;

    float timestep =
        (s->slowAnimations ? 2 : // For smooth slow-mo (refer to display.c)
         as->opt[ANIM_SCREEN_OPTION_TIME_STEP].value.i);

    model->remainderSteps += time / timestep;
    steps = floor (model->remainderSteps);
    model->remainderSteps -= steps;

    if (!steps && aw->animRemainingTime < aw->animTotalTime)
        return;
    steps = MAX(1, steps);

    Bool minimizeToTop = (s->height - aw->icon.y > aw->icon.y);

    for (j = 0; j < steps; j++)
    {
        float forwardProgress =
            1 - (aw->animRemainingTime - timestep) /
                (aw->animTotalTime - timestep);
        if (aw->curWindowEvent == WindowEventUnminimize)
            forwardProgress = 1 - forwardProgress;
        for (i = 0; i < model->numObjects; i++)
        {
            fxMagicLampModelStepObject (w, model,
                                        &model->objects[i],
                                        forwardProgress,
                                        minimizeToTop);
        }
        aw->animRemainingTime -= timestep;
        if (aw->animRemainingTime <= 0)
        {
            aw->animRemainingTime = 0; // avoid sub-zero values
            break;
        }
    }
    modelCalcBounds (model);
}


// =====================  Effect: Dream  =========================


static void
fxDreamInit (CompScreen * s, CompWindow * w)
{
    ANIM_WINDOW (w);
    ANIM_SCREEN (s);

    // store window opacity
    aw->storedOpacity = w->paint.opacity;

    aw->timestep =
        (s->slowAnimations ? 2 : // For smooth slow-mo (refer to display.c)
         as->opt[ANIM_SCREEN_OPTION_TIME_STEP].value.i);
}

static void
fxDreamModelStepObject (CompWindow * w,
                        Model * model, Object * object, float forwardProgress)
{
    float waveAmpMax = MIN (WIN_H (w), WIN_W (w)) * 0.125f;
    float waveWidth = 10.0f;
    float waveSpeed = 7.0f;

    float origx =
        WIN_X (w) +
        WIN_W (w) * object->gridPosition.x;
    float origy =
        WIN_Y (w) +
        WIN_H (w) * object->gridPosition.y;

    object->position.y = origy;
    object->position.x =
        origx +
        forwardProgress * waveAmpMax *
        sin (object->gridPosition.y * M_PI * waveWidth +
             waveSpeed * forwardProgress);

}

static void
fxDreamModelStep (CompScreen * s, CompWindow * w, float time)
{
    int i, j, steps;

    ANIM_WINDOW (w);
    ANIM_SCREEN (s);

    Model *model = aw->model;

    float timestep =
        (s->slowAnimations ? 2 : // For smooth slow-mo (refer to display.c)
         as->opt[ANIM_SCREEN_OPTION_TIME_STEP].value.i);
    aw->timestep = timestep;

    model->remainderSteps += time / timestep;
    steps = floor (model->remainderSteps);
    model->remainderSteps -= steps;

    if (!steps && aw->animRemainingTime < aw->animTotalTime)
        return;
    steps = MAX(1, steps);

    for (j = 0; j < steps; j++)
    {
        float forwardProgress =
            1 - (aw->animRemainingTime - timestep) /
                (aw->animTotalTime - timestep);
        if (aw->curWindowEvent == WindowEventCreate ||
            aw->curWindowEvent == WindowEventUnminimize)
            forwardProgress = 1 - forwardProgress;

        for (i = 0; i < model->numObjects; i++)
        {
            fxDreamModelStepObject (w,
                                    model,
                                    &model->objects[i], forwardProgress);
        }
        aw->animRemainingTime -= timestep;
        if (aw->animRemainingTime <= 0)
        {
            aw->animRemainingTime = 0; // avoid sub-zero values
            break;
        }
    }
    modelCalcBounds (model);
}

static void
fxDreamUpdateWindowAttrib (AnimScreen * as,
                           AnimWindow * aw,
                           WindowPaintAttrib * wAttrib)
{
    float forwardProgress =
        1 - (aw->animRemainingTime - aw->timestep) /
            (aw->animTotalTime - aw->timestep);
    forwardProgress = MIN (forwardProgress, 1);
    forwardProgress = MAX (forwardProgress, 0);

    if (aw->curWindowEvent == WindowEventCreate ||
        aw->curWindowEvent == WindowEventUnminimize)
        forwardProgress = 1 - forwardProgress;

    wAttrib->opacity = aw->storedOpacity * (1 - forwardProgress);
}


// =====================  Effect: Wave  =========================


static void
fxWaveModelStepObject (CompWindow * w,
                       Model * model,
                       Object * object,
                       float forwardProgress,
                       float waveAmp, float waveHalfWidth)
{
    float origx =
        w->attrib.x +
        (WIN_W (w) * object->gridPosition.x - w->output.left) * model->scale.x;
    float origy =
        w->attrib.y +
        (WIN_H (w) * object->gridPosition.y - w->output.top) * model->scale.y;

    float wavePosition =
        WIN_Y (w) - waveHalfWidth +
        forwardProgress * (WIN_H (w) * model->scale.y + 2 * waveHalfWidth);

    object->position.y = origy;
    object->position.x = origx;

    if (fabs (object->position.y - wavePosition) < waveHalfWidth)
        object->position.x +=
            (object->gridPosition.x - 0.5) * waveAmp *
            (cos ((object->position.y - wavePosition) * M_PI / waveHalfWidth)
             + 1) / 2;
}

static void
fxWaveModelStep (CompScreen * s, CompWindow * w, float time)
{
    int i, j, steps;

    ANIM_SCREEN (s);
    ANIM_WINDOW (w);

    Model *model = aw->model;

    float timestep =
        (s->slowAnimations ? 2 : // For smooth slow-mo (refer to display.c)
         as->opt[ANIM_SCREEN_OPTION_TIME_STEP].value.i);

    model->remainderSteps += time / timestep;
    steps = floor (model->remainderSteps);
    model->remainderSteps -= steps;

    if (!steps && aw->animRemainingTime < aw->animTotalTime)
        return;
    steps = MAX(1, steps);

    for (j = 0; j < steps; j++)
    {
        float forwardProgress =
            1 - (aw->animRemainingTime - timestep) /
                (aw->animTotalTime - timestep);

        for (i = 0; i < model->numObjects; i++)
        {
            fxWaveModelStepObject (w,
                                   model,
                                   &model->objects[i],
                                   forwardProgress,
                                   WIN_H (w) * model->scale.y *
                                   as->opt[ANIM_SCREEN_OPTION_FOCUS_WAVE_AMP].
                                   value.f,
                                   WIN_H (w) * model->scale.y *
                                   as->
                                   opt[ANIM_SCREEN_OPTION_FOCUS_WAVE_WIDTH].
                                   value.f / 2);
        }
        aw->animRemainingTime -= timestep;
        if (aw->animRemainingTime <= 0)
        {
            aw->animRemainingTime = 0; // avoid sub-zero values
            break;
        }
    }
    modelCalcBounds (model);
}


// =====================  Effect: Zoom and Sidekick  =========================


static void
fxSidekickInit (CompScreen * s, CompWindow * w)
{
    ANIM_WINDOW (w);
    ANIM_SCREEN (s);

    // determine number of rotations randomly in [0.75, 1.25] range
    aw->numZoomRotations = 1.0f + 0.5 * rand () / RAND_MAX - 0.25;

    // store window opacity
    aw->storedOpacity = w->paint.opacity;
    aw->timestep =
        (s->slowAnimations ? 2 : // For smooth slow-mo (refer to display.c)
         as->opt[ANIM_SCREEN_OPTION_TIME_STEP].value.i);
}

static void
fxZoomInit (CompScreen * s, CompWindow * w)
{
    ANIM_WINDOW (w);
    ANIM_SCREEN (s);

    // store window opacity
    aw->storedOpacity = w->paint.opacity;
    aw->timestep =
        (s->slowAnimations ? 2 : // For smooth slow-mo (refer to display.c)
         as->opt[ANIM_SCREEN_OPTION_TIME_STEP].value.i);
}

static float
fxZoomAnimProgressDir (AnimScreen * as, AnimWindow * aw)
{
    float forwardProgress =
        1 - (aw->animRemainingTime - aw->timestep) /
            (aw->animTotalTime - aw->timestep);

    int animProgressDir = 1;
    if (aw->curWindowEvent == WindowEventUnminimize ||
        aw->curWindowEvent == WindowEventCreate)
        animProgressDir = 2;

    if (aw->animOverrideProgressDir != 0)
        animProgressDir = aw->animOverrideProgressDir;

    float x = forwardProgress;

    if ((animProgressDir == 1 &&
         (aw->curWindowEvent == WindowEventUnminimize ||
          aw->curWindowEvent == WindowEventCreate)) ||
        (animProgressDir == 2 &&
         (aw->curWindowEvent == WindowEventMinimize ||
          aw->curWindowEvent == WindowEventClose)))
        x = 1 - forwardProgress;

    float m = 8;
    float minx = 0.9;
    float maxx = 1.3;

    forwardProgress = (1 - ((sigmoid2 (minx + (x*(maxx-minx)), m) - sigmoid2 (minx, m)) /
                            (sigmoid2 (maxx, m) - sigmoid2 (minx, m))));

    if (animProgressDir == 1)
        forwardProgress = 1 - forwardProgress;

    return forwardProgress;
}

static void
fxSidekickModelStepObject (CompWindow * w,
                           Model * model,
                           Object * object,
                           float forwardProgress,
                           Point rotateCenter, float sinRot, float cosRot)
{
    ANIM_WINDOW (w);

    float x =
        (1 - forwardProgress) *
        (WIN_X (w) +
         WIN_W (w) * object->gridPosition.x * model->scale.x) +
        forwardProgress * (aw->icon.x +
                           aw->icon.width * object->gridPosition.x);
    float y =
        (1 - forwardProgress) * (WIN_Y (w) +
                                 WIN_H (w) *
                                 object->gridPosition.y * model->scale.y) +
        forwardProgress * (aw->icon.y - 0 +
                           aw->icon.height * object->gridPosition.y);

    x -= rotateCenter.x;
    y -= rotateCenter.y;

    object->position.x = cosRot * x - sinRot * y;
    object->position.y = sinRot * x + cosRot * y;

    object->position.x += rotateCenter.x;
    object->position.y += rotateCenter.y;
}

static void
fxZoomModelStepObject (CompWindow * w,
                       Model * model, Object * object, float forwardProgress)
{
    ANIM_WINDOW (w);

    object->position.x =
        (1 - forwardProgress) *
        (WIN_X (w) +
         WIN_W (w) * object->gridPosition.x * model->scale.x) +
        forwardProgress * (aw->icon.x +
                           aw->icon.width * object->gridPosition.x);
    object->position.y =
        (1 - forwardProgress) * (WIN_Y (w) +
                                 WIN_H (w) *
                                 object->gridPosition.y * model->scale.y) +
        forwardProgress * (aw->icon.y - 0 +
                           aw->icon.height * object->gridPosition.y);
}

static void
fxZoomModelStep (CompScreen * s, CompWindow * w, float time)
{
    int i, j, steps;

    ANIM_SCREEN (s);
    ANIM_WINDOW (w);

    Model *model = aw->model;

    float timestep =
        (s->slowAnimations ? 2 : // For smooth slow-mo (refer to display.c)
         as->opt[ANIM_SCREEN_OPTION_TIME_STEP].value.i);
    aw->timestep = timestep;

    model->remainderSteps += time / timestep;
    steps = floor (model->remainderSteps);
    model->remainderSteps -= steps;

    if (!steps && aw->animRemainingTime < aw->animTotalTime)
        return;
    steps = MAX(1, steps);

    for (j = 0; j < steps; j++)
    {
        float forwardProgress = fxZoomAnimProgressDir (as, aw);
        float sinRot = 0;
        float cosRot = 0;

        Point rotateCenter = {0, 0};

        if (aw->curAnimEffect == AnimEffectSidekick)
        {
            rotateCenter.x =
                (1 - forwardProgress) * (WIN_X (w) +
                                         WIN_W (w) * model->scale.x / 2) +
                forwardProgress * (aw->icon.x + aw->icon.width / 2);
            rotateCenter.y =
                (1 - forwardProgress) * (WIN_Y (w) +
                                         WIN_H (w) * model->scale.y / 2) +
                forwardProgress * (aw->icon.y + aw->icon.height / 2);

            sinRot = sin (2 * M_PI * forwardProgress * aw->numZoomRotations);
            cosRot = cos (2 * M_PI * forwardProgress * aw->numZoomRotations);
        }

        if (aw->curAnimEffect == AnimEffectSidekick)
            for (i = 0; i < model->numObjects; i++)
                fxSidekickModelStepObject (w, model, &model->objects[i],
                                           forwardProgress, rotateCenter, sinRot, cosRot);
        else
            for (i = 0; i < model->numObjects; i++)
                fxZoomModelStepObject (w, model, &model->objects[i],
                                       forwardProgress);

        aw->animRemainingTime -= timestep;
        if (aw->animRemainingTime <= 0)
        {
            aw->animRemainingTime = 0; // avoid sub-zero values
            break;
        }
    }
    modelCalcBounds (model);
}

static void
fxZoomUpdateWindowAttrib (AnimScreen * as,
                          AnimWindow * aw,
                          WindowPaintAttrib * wAttrib)
{
    float forwardProgress = fxZoomAnimProgressDir (as, aw);
    forwardProgress = MIN (forwardProgress, 1);
    forwardProgress = MAX (forwardProgress, 0);

    wAttrib->opacity = aw->storedOpacity * pow (1 - forwardProgress, 0.5);
}


// =====================  Effect: Curved Fold  =========================


static void
fxCurvedFoldModelStepObject (CompWindow * w,
                             Model * model,
                             Object * object,
                             float forwardProgress,
                             float curveMaxAmp)
{
    ANIM_WINDOW (w);

    float origx =
        WIN_X (w) +
        WIN_W (w) * object->gridPosition.x;

    if (aw->curWindowEvent == WindowEventShade ||
        aw->curWindowEvent == WindowEventUnshade)
    {
        // Execute shade mode

        // find position in window contents
        // (window contents correspond to 0.0-1.0 range)
        float relPosInWinContents =
            (object->gridPosition.y * WIN_H (w) - model->topHeight) /
            w->height;
        float relDistToCenter = fabs (relPosInWinContents - 0.5);

        if (object->gridPosition.y == 0)
        {
            object->position.x = origx;
            object->position.y = WIN_Y (w);
        }
        else if (object->gridPosition.y == 1)
        {
            object->position.x = origx;
            object->position.y =
                (1 - forwardProgress) *
                (WIN_Y (w) + WIN_H (w) * object->gridPosition.y) +
                forwardProgress *
                (WIN_Y (w) + model->topHeight + model->bottomHeight);
        }
        else
        {
            object->position.x =
                origx + sin (forwardProgress * M_PI / 2) *
                        (0.5 - object->gridPosition.x) * 2 *
                        (curveMaxAmp - curveMaxAmp * 4 * relDistToCenter*relDistToCenter);
            object->position.y =
                (1 - forwardProgress) *
                (WIN_Y (w) + WIN_H (w) * object->gridPosition.y) +
                forwardProgress *
                (WIN_Y (w) + model->topHeight);
        }
    }
    else
    {   // Execute normal mode

        // find position within window borders
        // (border contents correspond to 0.0-1.0 range)
        float relPosInWinBorders =
            (object->gridPosition.y * WIN_H (w) - (w->output.top - w->input.top)) /
            BORDER_H (w);
        float relDistToCenter = fabs (relPosInWinBorders - 0.5);

        // prevent top & bottom shadows from extending too much
        if (relDistToCenter > 0.5)
            relDistToCenter = 0.5;

        object->position.x =
            origx + sin (forwardProgress * M_PI / 2) *
                    (0.5 - object->gridPosition.x) * 2 *
                    (curveMaxAmp - curveMaxAmp * 4 * relDistToCenter*relDistToCenter);
        object->position.y =
            (1 - forwardProgress) *
            (WIN_Y (w) + WIN_H (w) * object->gridPosition.y) +
            forwardProgress *
            BORDER_Y (w);
    }
}

static void
fxCurvedFoldModelStep (CompScreen * s, CompWindow * w, float time)
{
    int i, j, steps;

    ANIM_SCREEN (s);
    ANIM_WINDOW (w);

    Model *model = aw->model;

    float timestep =
        (s->slowAnimations ? 2 : // For smooth slow-mo (refer to display.c)
         as->opt[ANIM_SCREEN_OPTION_TIME_STEP].value.i);

    model->remainderSteps += time / timestep;
    steps = floor (model->remainderSteps);
    model->remainderSteps -= steps;

    if (!steps && aw->animRemainingTime < aw->animTotalTime)
        return;
    steps = MAX(1, steps);

    for (j = 0; j < steps; j++)
    {
        float forwardProgress =
            1 - (aw->animRemainingTime - timestep) /
                (aw->animTotalTime - timestep);
        if (aw->curWindowEvent == WindowEventCreate ||
            aw->curWindowEvent == WindowEventUnminimize ||
            aw->curWindowEvent == WindowEventUnshade)
            forwardProgress = 1 - forwardProgress;

        for (i = 0; i < model->numObjects; i++)
            fxCurvedFoldModelStepObject (
                 w, model, &model->objects[i],
                 forwardProgress,
                 as->opt[ANIM_SCREEN_OPTION_CURVED_FOLD_AMP].value.f *
                 WIN_W (w));

        aw->animRemainingTime -= timestep;
        if (aw->animRemainingTime <= 0)
        {
            aw->animRemainingTime = 0; // avoid sub-zero values
            break;
        }
    }
    modelCalcBounds (model);
}


// =====================  Effect: Horizontal Folds  =========================


static void
fxHorizontalFoldsInitGrid (AnimScreen * as,
                           WindowEvent forWindowEvent,
                           int * gridWidth,
                           int * gridHeight)
{
    *gridWidth = 2;
    if (forWindowEvent == WindowEventShade ||
        forWindowEvent == WindowEventUnshade)
        *gridHeight =
            3 + 2 *
            as->opt[ANIM_SCREEN_OPTION_HORIZONTAL_FOLDS_NUM_FOLDS].value.i;
    else
        *gridHeight =
            1 + 2 *
            as->opt[ANIM_SCREEN_OPTION_HORIZONTAL_FOLDS_NUM_FOLDS].value.i;
}

static void
fxHorizontalFoldsModelStepObject (CompWindow * w,
                                  Model * model,
                                  Object * object,
                                  float forwardProgress,
                                  float curveMaxAmp,
                                  int rowNo)
{
    ANIM_WINDOW (w);

    float origx =
        WIN_X (w) +
        WIN_W (w) * object->gridPosition.x;

    if (aw->curWindowEvent == WindowEventShade ||
        aw->curWindowEvent == WindowEventUnshade)
    {
        // Execute shade mode

        float relDistToFoldCenter = (rowNo % 2 == 1 ? 0.5 : 0);

        if (object->gridPosition.y == 0)
        {
            object->position.x = origx;
            object->position.y = WIN_Y (w);
        }
        else if (object->gridPosition.y == 1)
        {
            object->position.x = origx;
            object->position.y =
                (1 - forwardProgress) *
                (WIN_Y (w) + WIN_H (w) * object->gridPosition.y) +
                forwardProgress *
                (WIN_Y (w) + model->topHeight + model->bottomHeight);
        }
        else
        {
            object->position.x =
                origx + sin (forwardProgress * M_PI / 2) *
                        (0.5 - object->gridPosition.x) * 2 *
                        (curveMaxAmp - curveMaxAmp * 4 *
                         relDistToFoldCenter * relDistToFoldCenter);
            object->position.y =
                (1 - forwardProgress) *
                (WIN_Y (w) + WIN_H (w) * object->gridPosition.y) +
                forwardProgress *
                (WIN_Y (w) + model->topHeight);
        }
    }
    else
    {   // Execute normal mode

        float relDistToFoldCenter;
        relDistToFoldCenter = (rowNo % 2 == 0 ? 0.5 : 0);

        object->position.x =
            origx + sin (forwardProgress * M_PI / 2) *
                    (0.5 - object->gridPosition.x) * 2 *
                    (curveMaxAmp - curveMaxAmp * 4 *
                     relDistToFoldCenter * relDistToFoldCenter);
        object->position.y =
            (1 - forwardProgress) *
            (WIN_Y (w) + WIN_H (w) * object->gridPosition.y) +
            forwardProgress *
            BORDER_Y (w);
    }
}

static void
fxHorizontalFoldsModelStep (CompScreen * s, CompWindow * w, float time)
{
    int i, j, steps;

    ANIM_SCREEN (s);
    ANIM_WINDOW (w);

    Model *model = aw->model;

    float timestep =
        (s->slowAnimations ? 2 : // For smooth slow-mo (refer to display.c)
         as->opt[ANIM_SCREEN_OPTION_TIME_STEP].value.i);

    model->remainderSteps += time / timestep;
    steps = floor (model->remainderSteps);
    model->remainderSteps -= steps;
    if (!steps && aw->animRemainingTime < aw->animTotalTime)
        return;
    steps = MAX(1, steps);

    for (j = 0; j < steps; j++)
    {
        float forwardProgress =
            1 - (aw->animRemainingTime - timestep) /
                (aw->animTotalTime - timestep);
        if (aw->curWindowEvent == WindowEventCreate ||
            aw->curWindowEvent == WindowEventUnminimize ||
            aw->curWindowEvent == WindowEventUnshade)
            forwardProgress = 1 - forwardProgress;

        for (i = 0; i < model->numObjects; i++)
            fxHorizontalFoldsModelStepObject (
                 w, model, &model->objects[i],
                 forwardProgress,
                 as->opt[ANIM_SCREEN_OPTION_HORIZONTAL_FOLDS_AMP].value.f *
                 WIN_W (w),
                 i / model->gridWidth);

        aw->animRemainingTime -= timestep;
        if (aw->animRemainingTime <= 0)
        {
            aw->animRemainingTime = 0; // avoid sub-zero values
            break;
        }
    }
    modelCalcBounds (model);
}



// =====================  Effect: Roll Up  =========================


static void
fxRollUpInitGrid (AnimScreen * as,
                  WindowEvent forWindowEvent,
                  int * gridWidth,
                  int * gridHeight)
{
    *gridWidth = 2;
    if (forWindowEvent == WindowEventShade ||
        forWindowEvent == WindowEventUnshade)
        *gridHeight = 4; 
    else
        *gridHeight = 2; 
}

static void
fxRollUpModelStepObject (CompWindow * w,
                         Model * model,
                         Object * object,
                         float forwardProgress,
                         Bool fixedInterior)
{
    ANIM_WINDOW (w);

    float origx =
        WIN_X (w) +
        WIN_W (w) * object->gridPosition.x;

    if (aw->curWindowEvent == WindowEventShade ||
        aw->curWindowEvent == WindowEventUnshade)
    {
        // Execute shade mode

        // find position in window contents
        // (window contents correspond to 0.0-1.0 range)
        float relPosInWinContents =
            (object->gridPosition.y * WIN_H (w) - model->topHeight) /
            w->height;

        if (object->gridPosition.y == 0)
        {
            object->position.x = origx;
            object->position.y = WIN_Y (w);
        }
        else if (object->gridPosition.y == 1)
        {
            object->position.x = origx;
            object->position.y =
                (1 - forwardProgress) *
                (WIN_Y (w) + WIN_H (w) * object->gridPosition.y) +
                forwardProgress *
                (WIN_Y (w) + model->topHeight + model->bottomHeight);
        }
        else
        {
            object->position.x = origx;

            if (relPosInWinContents > forwardProgress)
            {
                object->position.y =
                    (1 - forwardProgress) *
                    (WIN_Y (w) + WIN_H (w) * object->gridPosition.y) +
                    forwardProgress *
                    (WIN_Y (w) + model->topHeight);

                if (fixedInterior)
                    object->offsetTexCoordForQuadBefore.y =
                        - forwardProgress * w->height;
            }
            else
            {
                object->position.y = WIN_Y (w) + model->topHeight;
                if (! fixedInterior)
                    object->offsetTexCoordForQuadAfter.y =
                        (forwardProgress - relPosInWinContents) * w->height;
            }
        }
    }
    /*else            Doesn't work yet
    {   // Execute normal mode

        // find position within window borders
        // (border contents correspond to 0.0-1.0 range)
        //float relPosInWinBorders =
        //    (object->gridPosition.y * WIN_H (w) - (w->output.top - w->input.top)) /
        //    BORDER_H (w);

        object->position.x = origx;

        if (object->gridPosition.y > forwardProgress)
        {
            object->position.y =
                (1 - forwardProgress) *
                (WIN_Y (w) + WIN_H (w) * object->gridPosition.y) +
                forwardProgress *
                WIN_Y (w);
            //printf("1 obj.position.y: %.2f, relPosInWinContents: %.2f, forwardProgress: %.2f\n",
            //       object->position.y, relPosInWinContents, forwardProgress);
        }
        else
        {
            object->position.y = WIN_Y (w);
            object->offsetTexCoordForQuadAfter.y =
                (forwardProgress - object->gridPosition.y) * WIN_H (w);
            //printf("2 obj.position.y: %.2f, relPosInWinContents: %.2f, forwardProgress: %.2f\n",
            //       object->position.y, relPosInWinContents, forwardProgress);
        }
    }*/
}

static void
fxRollUpModelStep (CompScreen * s, CompWindow * w, float time)
{
    int i, j, steps;

    ANIM_SCREEN (s);
    ANIM_WINDOW (w);

    Model *model = aw->model;

    float timestep =
        (s->slowAnimations ? 2 : // For smooth slow-mo (refer to display.c)
         as->opt[ANIM_SCREEN_OPTION_TIME_STEP].value.i);

    model->remainderSteps += time / timestep;
    steps = floor (model->remainderSteps);
    model->remainderSteps -= steps;
    if (!steps && aw->animRemainingTime < aw->animTotalTime)
        return;
    steps = MAX(1, steps);

    for (j = 0; j < steps; j++)
    {
        float forwardProgress =
            1 - (aw->animRemainingTime - timestep) /
                (aw->animTotalTime - timestep);
        if (aw->curWindowEvent == WindowEventCreate ||
            aw->curWindowEvent == WindowEventUnminimize ||
            aw->curWindowEvent == WindowEventUnshade)
            forwardProgress = 1 - forwardProgress;

        for (i = 0; i < model->numObjects; i++)
            fxRollUpModelStepObject (
                w, model, &model->objects[i],
                forwardProgress,
                as->opt[ANIM_SCREEN_OPTION_ROLLUP_FIXED_INTERIOR].value.b);

        aw->animRemainingTime -= timestep;
        if (aw->animRemainingTime <= 0)
        {
            aw->animRemainingTime = 0; // avoid sub-zero values
            break;
        }
    }
    modelCalcBounds (model);
}


// =====================  Effect: Fade  =========================


static void
fxFadeInit (CompScreen * s, CompWindow * w)
{
    ANIM_WINDOW (w);
    ANIM_SCREEN (s);

    // store window opacity
    aw->storedOpacity = w->paint.opacity;

    aw->timestep =
        (s->slowAnimations ? 2 : // For smooth slow-mo (refer to display.c)
         as->opt[ANIM_SCREEN_OPTION_TIME_STEP].value.i);
}

static void
fxFadeModelStep (CompScreen * s, CompWindow * w, float time)
{
    int j, steps;

    ANIM_WINDOW (w);
    ANIM_SCREEN (s);

    Model *model = aw->model;

    float timestep =
        (s->slowAnimations ? 2 : // For smooth slow-mo (refer to display.c)
         as->opt[ANIM_SCREEN_OPTION_TIME_STEP].value.i);
    aw->timestep = timestep;

    model->remainderSteps += time / timestep;
    steps = floor (model->remainderSteps);
    model->remainderSteps -= steps;

    if (!steps && aw->animRemainingTime < aw->animTotalTime)
        return;
    steps = MAX(1, steps);

    for (j = 0; j < steps; j++)
    {
        aw->animRemainingTime -= timestep;
        if (aw->animRemainingTime <= 0)
        {
            aw->animRemainingTime = 0; // avoid sub-zero values
            break;
        }
    }
}

static void
fxFadeUpdateWindowAttrib (AnimScreen * as,
                           AnimWindow * aw,
                           WindowPaintAttrib * wAttrib)
{
    float forwardProgress =
        1 - (aw->animRemainingTime - aw->timestep) /
            (aw->animTotalTime - aw->timestep);
    forwardProgress = MIN (forwardProgress, 1);
    forwardProgress = MAX (forwardProgress, 0);

    if (aw->curWindowEvent == WindowEventCreate ||
        aw->curWindowEvent == WindowEventUnminimize)
        forwardProgress = 1 - forwardProgress;

    wAttrib->opacity = aw->storedOpacity * (1 - forwardProgress);
}


// =====================  End of Effects  =========================


typedef struct _AnimEffectProperties
{
    void (*updateWindowAttribFunc) (
        AnimScreen *, AnimWindow *, WindowPaintAttrib *);
    void (*preDrawGeometryFunc) (AnimWindow *);
    void (*postDrawGeometryFunc) (AnimWindow *);
    void (*modelStepFunc) (CompScreen *, CompWindow *, float time);
    void (*initFunc) (CompScreen *, CompWindow *);
    void (*initGridFunc) (AnimScreen *, WindowEvent, int *, int *);
} AnimEffectProperties;

AnimEffectProperties animEffectProperties[AnimEffectNum] = {
    // AnimEffectNone
    {0, 0, 0, 0, 0, 0},
    // AnimEffectFade
    {fxFadeUpdateWindowAttrib, 0, 0, fxFadeModelStep, fxFadeInit, 0},
    // AnimEffectMagicLamp1
    {0, 0, 0, fxMagicLampModelStep, fxMagicLampInit, fxMagicLamp1InitGrid},
    // AnimEffectMagicLamp2
    {0, 0, 0, fxMagicLampModelStep, fxMagicLampInit, fxMagicLamp2InitGrid},
    // AnimEffectZoom
    {fxZoomUpdateWindowAttrib, 0, 0, fxZoomModelStep, fxZoomInit, 0},
    // AnimEffectSidekick
    {fxZoomUpdateWindowAttrib, 0, 0, fxZoomModelStep, fxSidekickInit, 0},
    // AnimEffectDream
    {fxDreamUpdateWindowAttrib, 0, 0, fxDreamModelStep, fxDreamInit, fxMagicLamp1InitGrid},
    // AnimEffectWave
    {0, 0, 0, fxWaveModelStep, 0, fxMagicLamp1InitGrid},
    // AnimEffectCurvedFold
    {0, 0, 0, fxCurvedFoldModelStep, 0, fxMagicLamp1InitGrid},
    // AnimEffectHorizontalFolds
    {0, 0, 0, fxHorizontalFoldsModelStep, 0, fxHorizontalFoldsInitGrid},
    // AnimEffectRollUp
    {0, 0, 0, fxRollUpModelStep, 0, fxRollUpInitGrid}
};


static Bool
getMousePointerXY (CompScreen * s, short *x, short *y)
{
    Window w1, w2;
    int xp, yp, xj, yj;
    unsigned int m;
    if (XQueryPointer
        (s->display->display, s->root, &w1, &w2, &xj, &yj, &xp, &yp, &m))
    {
        *x = xp;
        *y = yp;
        return TRUE;
    }
    return FALSE;
}

static int
animGetWindowState (CompWindow * w)
{
    Atom actual;
    int result, format;
    unsigned long n, left;
    unsigned char *data;

    result = XGetWindowProperty (w->screen->display->display, w->id,
                                 w->screen->display->wmStateAtom, 0L, 1L,
                                 FALSE, w->screen->display->wmStateAtom,
                                 &actual, &format, &n, &left, &data);

    if (result == Success && n && data)
    {
        int state;

        memcpy (&state, data, sizeof (int));
        XFree ((void *)data);

        return state;
    }

    return WithdrawnState;
}

static Bool
animSetScreenOption (CompScreen * screen, char *name, CompOptionValue * value)
{
    CompOption *o;
    int index;

    ANIM_SCREEN (screen);

    o = compFindOption (as->opt, NUM_OPTIONS (as), name, &index);
    if (!o)
        return FALSE;

    switch (index)
    {
    case ANIM_SCREEN_OPTION_ZOOM_FROM_CENTER:
        if (compSetBoolOption (o, value))
            return TRUE;
        break;
    case ANIM_SCREEN_OPTION_TIME_STEP:
        if (compSetIntOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_DISABLE_PP_FX:
        if (compSetStringOption (o, value))
        {
            int i;

            for (i = 0; i < 3; i++)
            {
                if (strcmp (o->value.s, ppDisablingName[i]) == 0)
                {
                    as->ppDisabling = i;
                    return TRUE;
                }
            }
        }
        break;
    case ANIM_SCREEN_OPTION_MAGIC_LAMP1_GRID_RES:
        if (compSetIntOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_MAGIC_LAMP1_MAX_WAVES:
        if (compSetIntOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_MAGIC_LAMP1_WAVE_AMP_MIN:
        if (compSetFloatOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_MAGIC_LAMP1_WAVE_AMP_MAX:
        if (compSetFloatOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_MAGIC_LAMP2_GRID_RES:
        if (compSetIntOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_MAGIC_LAMP2_MAX_WAVES:
        if (compSetIntOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_MAGIC_LAMP2_WAVE_AMP_MIN:
        if (compSetFloatOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_MAGIC_LAMP2_WAVE_AMP_MAX:
        if (compSetFloatOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_FOCUS_WAVE_WIDTH:
        if (compSetFloatOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_FOCUS_WAVE_AMP:
        if (compSetFloatOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_CURVED_FOLD_AMP:
        if (compSetFloatOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_HORIZONTAL_FOLDS_AMP:
        if (compSetFloatOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_HORIZONTAL_FOLDS_NUM_FOLDS:
        if (compSetIntOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_ROLLUP_FIXED_INTERIOR:
        if (compSetBoolOption (o, value))
            return TRUE;
        break;
    case ANIM_SCREEN_OPTION_MINIMIZE_EFFECT:
        if (compSetStringOption (o, value))
        {
            int i;

            for (i = 0; i < NUM_MINIMIZE_EFFECT; i++)
            {
                if (strcmp (o->value.s, minimizeEffectName[i]) == 0)
                {
                    as->minimizeEffect = minimizeEffectType[i];
                    return TRUE;
                }
            }
        }
        break;
    case ANIM_SCREEN_OPTION_UNMINIMIZE_EFFECT:
        if (compSetStringOption (o, value))
        {
            int i;

            for (i = 0; i < NUM_MINIMIZE_EFFECT; i++)
            {
                if (strcmp (o->value.s, minimizeEffectName[i]) == 0)
                {
                    as->unminimizeEffect = minimizeEffectType[i];
                    return TRUE;
                }
            }
        }
        break;
    case ANIM_SCREEN_OPTION_CLOSE1_EFFECT:
        if (compSetStringOption (o, value))
        {
            int i;

            for (i = 0; i < NUM_CLOSE_EFFECT; i++)
            {
                if (strcmp (o->value.s, closeEffectName[i]) == 0)
                {
                    as->close1Effect = closeEffectType[i];
                    return TRUE;
                }
            }
        }
        break;
    case ANIM_SCREEN_OPTION_CLOSE2_EFFECT:
        if (compSetStringOption (o, value))
        {
            int i;

            for (i = 0; i < NUM_CLOSE_EFFECT; i++)
            {
                if (strcmp (o->value.s, closeEffectName[i]) == 0)
                {
                    as->close2Effect = closeEffectType[i];
                    return TRUE;
                }
            }
        }
        break;
    case ANIM_SCREEN_OPTION_CREATE1_EFFECT:
        if (compSetStringOption (o, value))
        {
            int i;

            for (i = 0; i < NUM_CLOSE_EFFECT; i++)
            {
                if (strcmp (o->value.s, closeEffectName[i]) == 0)
                {
                    as->create1Effect = closeEffectType[i];
                    return TRUE;
                }
            }
        }
        break;
    case ANIM_SCREEN_OPTION_CREATE2_EFFECT:
        if (compSetStringOption (o, value))
        {
            int i;

            for (i = 0; i < NUM_CLOSE_EFFECT; i++)
            {
                if (strcmp (o->value.s, closeEffectName[i]) == 0)
                {
                    as->create2Effect = closeEffectType[i];
                    return TRUE;
                }
            }
        }
        break;
    case ANIM_SCREEN_OPTION_FOCUS_EFFECT:
        if (compSetStringOption (o, value))
        {
            int i;

            for (i = 0; i < NUM_FOCUS_EFFECT; i++)
            {
                if (strcmp (o->value.s, focusEffectName[i]) == 0)
                {
                    as->focusEffect = focusEffectType[i];
                    return TRUE;
                }
            }
        }
        break;
    case ANIM_SCREEN_OPTION_SHADE_EFFECT:
        if (compSetStringOption (o, value))
        {
            int i;

            for (i = 0; i < NUM_SHADE_EFFECT; i++)
            {
                if (strcmp (o->value.s, shadeEffectName[i]) == 0)
                {
                    as->shadeEffect = shadeEffectType[i];
                    return TRUE;
                }
            }
        }
        break;
    case ANIM_SCREEN_OPTION_UNSHADE_EFFECT:
        if (compSetStringOption (o, value))
        {
            int i;

            for (i = 0; i < NUM_SHADE_EFFECT; i++)
            {
                if (strcmp (o->value.s, shadeEffectName[i]) == 0)
                {
                    as->unshadeEffect = shadeEffectType[i];
                    return TRUE;
                }
            }
        }
        break;
    case ANIM_SCREEN_OPTION_MINIMIZE_WINDOW_TYPE:
        if (compSetOptionList (o, value))
        {
            as->minimizeWMask = compWindowTypeMaskFromStringList (&o->value);
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_UNMINIMIZE_WINDOW_TYPE:
        if (compSetOptionList (o, value))
        {
            as->unminimizeWMask =
                compWindowTypeMaskFromStringList (&o->value);
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_CLOSE1_WINDOW_TYPE:
        if (compSetOptionList (o, value))
        {
            as->close1WMask = compWindowTypeMaskFromStringList (&o->value);
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_CLOSE2_WINDOW_TYPE:
        if (compSetOptionList (o, value))
        {
            as->close2WMask = compWindowTypeMaskFromStringList (&o->value);
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_CREATE1_WINDOW_TYPE:
        if (compSetOptionList (o, value))
        {
            as->create1WMask = compWindowTypeMaskFromStringList (&o->value);
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_CREATE2_WINDOW_TYPE:
        if (compSetOptionList (o, value))
        {
            as->create2WMask = compWindowTypeMaskFromStringList (&o->value);
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_FOCUS_WINDOW_TYPE:
        if (compSetOptionList (o, value))
        {
            as->focusWMask = compWindowTypeMaskFromStringList (&o->value);
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_SHADE_WINDOW_TYPE:
        if (compSetOptionList (o, value))
        {
            as->shadeWMask = compWindowTypeMaskFromStringList (&o->value);
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_UNSHADE_WINDOW_TYPE:
        if (compSetOptionList (o, value))
        {
            as->unshadeWMask = compWindowTypeMaskFromStringList (&o->value);
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_MINIMIZE_DURATION:
        if (compSetFloatOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_UNMINIMIZE_DURATION:
        if (compSetFloatOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_CLOSE1_DURATION:
        if (compSetFloatOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_CLOSE2_DURATION:
        if (compSetFloatOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_CREATE1_DURATION:
        if (compSetFloatOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_CREATE2_DURATION:
        if (compSetFloatOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_FOCUS_DURATION:
        if (compSetFloatOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_SHADE_DURATION:
        if (compSetFloatOption (o, value))
        {
            return TRUE;
        }
        break;
    case ANIM_SCREEN_OPTION_UNSHADE_DURATION:
        if (compSetFloatOption (o, value))
        {
            return TRUE;
        }
        break;
    default:
        break;
    }

    return FALSE;
}

static void
animScreenInitOptions (AnimScreen * as)
{
    CompOption *o;
    int i;

    o = &as->opt[ANIM_SCREEN_OPTION_ROLLUP_FIXED_INTERIOR];
    o->name = "rollup_fixed_interior";
    o->shortDesc = N_("Rollup Fixed Interior");
    o->longDesc = N_("Fixed window interior during Rollup effect");
    o->type = CompOptionTypeBool;
    o->value.b = ANIM_ROLLUP_FIXED_INTERIOR_DEFAULT;

    o = &as->opt[ANIM_SCREEN_OPTION_HORIZONTAL_FOLDS_NUM_FOLDS];
    o->name = "horizontal_folds_num_folds";
    o->shortDesc = N_("Horizontal Folds Number of Folds");
    o->longDesc = N_("Horizontal Folds number of complete folds");
    o->type = CompOptionTypeInt;
    o->value.i = ANIM_HORIZONTAL_FOLDS_NUM_FOLDS_DEFAULT;
    o->rest.i.min = ANIM_HORIZONTAL_FOLDS_NUM_FOLDS_MIN;
    o->rest.i.max = ANIM_HORIZONTAL_FOLDS_NUM_FOLDS_MAX;

    o = &as->opt[ANIM_SCREEN_OPTION_HORIZONTAL_FOLDS_AMP];
    o->name = "horizontal_folds_amp";
    o->shortDesc = N_("Horizontal Folds Amplitude");
    o->longDesc = N_("Horizontal folds amplitude relative to window width. Negative values fold outward.");
    o->type = CompOptionTypeFloat;
    o->value.f = ANIM_HORIZONTAL_FOLDS_AMP_DEFAULT;
    o->rest.f.min = ANIM_HORIZONTAL_FOLDS_AMP_MIN;
    o->rest.f.max = ANIM_HORIZONTAL_FOLDS_AMP_MAX;
    o->rest.f.precision = ANIM_HORIZONTAL_FOLDS_AMP_PRECISION;

    o = &as->opt[ANIM_SCREEN_OPTION_CURVED_FOLD_AMP];
    o->name = "curved_fold_amp";
    o->shortDesc = N_("Curved Fold Amplitude");
    o->longDesc = N_("Curved fold amplitude relative to window width. Negative values fold outward.");
    o->type = CompOptionTypeFloat;
    o->value.f = ANIM_CURVED_FOLD_AMP_DEFAULT;
    o->rest.f.min = ANIM_CURVED_FOLD_AMP_MIN;
    o->rest.f.max = ANIM_CURVED_FOLD_AMP_MAX;
    o->rest.f.precision = ANIM_CURVED_FOLD_AMP_PRECISION;

    o = &as->opt[ANIM_SCREEN_OPTION_ZOOM_FROM_CENTER];
    o->name = "zoom_from_center";
    o->shortDesc = N_("Zoom From Center");
    o->longDesc =
        N_("Zoom from center when playing Zoom and Sidekick effects");
    o->type = CompOptionTypeBool;
    o->value.b = ANIM_ZOOM_FROM_CENTER_DEFAULT;

    o = &as->opt[ANIM_SCREEN_OPTION_TIME_STEP];
    o->name = "time_step";
    o->shortDesc = N_("Animation Time Step");
    o->longDesc = N_("Animation time step in milliseconds");
    o->type = CompOptionTypeInt;
    o->value.i = ANIM_TIME_STEP_DEFAULT;
    o->rest.i.min = ANIM_TIME_STEP_MIN;
    o->rest.i.max = ANIM_TIME_STEP_MAX;

    o = &as->opt[ANIM_SCREEN_OPTION_DISABLE_PP_FX];
    o->name = "disable_pp_fx";
    o->shortDesc = N_("Disable Postprocessing Effects During Animation");
    o->longDesc = N_("Disables Blur during animation");
    o->type = CompOptionTypeString;
    o->value.s = strdup (ppDisablingName[ANIM_DISABLE_PP_FX_DEFAULT]);
    o->rest.s.string = ppDisablingName;
    o->rest.s.nString = 3;

    o = &as->opt[ANIM_SCREEN_OPTION_MAGIC_LAMP1_GRID_RES];
    o->name = "magic_lamp1_grid_res";
    o->shortDesc = N_("Magic Lamp 1 Grid Y Resolution");
    o->longDesc =
        N_
        ("Vertex Grid Resolution (2-200) for Magic Lamp 1 (Y dimension only)");
    o->type = CompOptionTypeInt;
    o->value.i = ANIM_MAGIC_LAMP1_GRID_RES_DEFAULT;
    o->rest.i.min = ANIM_MAGIC_LAMP1_GRID_RES_MIN;
    o->rest.i.max = ANIM_MAGIC_LAMP1_GRID_RES_MAX;

    o = &as->opt[ANIM_SCREEN_OPTION_MAGIC_LAMP1_MAX_WAVES];
    o->name = "magic_lamp1_max_waves";
    o->shortDesc = N_("Magic Lamp 1 Max Waves");
    o->longDesc = N_("Maximum number of waves for Magic Lamp 1");
    o->type = CompOptionTypeInt;
    o->value.i = ANIM_MAGIC_LAMP1_MAX_WAVES_DEFAULT;
    o->rest.i.min = ANIM_MAGIC_LAMP1_MAX_WAVES_MIN;
    o->rest.i.max = ANIM_MAGIC_LAMP1_MAX_WAVES_MAX;

    o = &as->opt[ANIM_SCREEN_OPTION_MAGIC_LAMP1_WAVE_AMP_MIN];
    o->name = "magic_lamp1_wave_amp_min";
    o->shortDesc = N_("Magic Lamp 1 Wave Min Amplitude");
    o->longDesc =
        N_("Magic Lamp 1 waves will have random amplitudes between min and max");
    o->type = CompOptionTypeFloat;
    o->value.f = ANIM_MAGIC_LAMP1_WAVE_AMP_MIN_DEFAULT;
    o->rest.f.min = ANIM_MAGIC_LAMP1_WAVE_AMP_MIN_MIN;
    o->rest.f.max = ANIM_MAGIC_LAMP1_WAVE_AMP_MIN_MAX;
    o->rest.f.precision = ANIM_MAGIC_LAMP1_WAVE_AMP_MIN_PRECISION;

    o = &as->opt[ANIM_SCREEN_OPTION_MAGIC_LAMP1_WAVE_AMP_MAX];
    o->name = "magic_lamp1_wave_amp_max";
    o->shortDesc = N_("Magic Lamp 1 Wave Max Amplitude");
    o->longDesc =
        N_("Magic Lamp 1 waves will have random amplitudes between min and max");
    o->type = CompOptionTypeFloat;
    o->value.f = ANIM_MAGIC_LAMP1_WAVE_AMP_MAX_DEFAULT;
    o->rest.f.min = ANIM_MAGIC_LAMP1_WAVE_AMP_MAX_MIN;
    o->rest.f.max = ANIM_MAGIC_LAMP1_WAVE_AMP_MAX_MAX;
    o->rest.f.precision = ANIM_MAGIC_LAMP1_WAVE_AMP_MAX_PRECISION;

    o = &as->opt[ANIM_SCREEN_OPTION_MAGIC_LAMP2_GRID_RES];
    o->name = "magic_lamp2_grid_res";
    o->shortDesc = N_("Magic Lamp 2 Grid Y Resolution");
    o->longDesc =
        N_
        ("Vertex Grid Resolution (2-200) for Magic Lamp 2 (Y dimension only)");
    o->type = CompOptionTypeInt;
    o->value.i = ANIM_MAGIC_LAMP2_GRID_RES_DEFAULT;
    o->rest.i.min = ANIM_MAGIC_LAMP2_GRID_RES_MIN;
    o->rest.i.max = ANIM_MAGIC_LAMP2_GRID_RES_MAX;

    o = &as->opt[ANIM_SCREEN_OPTION_MAGIC_LAMP2_MAX_WAVES];
    o->name = "magic_lamp2_max_waves";
    o->shortDesc = N_("Magic Lamp 2 Max Waves");
    o->longDesc = N_("Maximum number of waves for Magic Lamp 2");
    o->type = CompOptionTypeInt;
    o->value.i = ANIM_MAGIC_LAMP2_MAX_WAVES_DEFAULT;
    o->rest.i.min = ANIM_MAGIC_LAMP2_MAX_WAVES_MIN;
    o->rest.i.max = ANIM_MAGIC_LAMP2_MAX_WAVES_MAX;

    o = &as->opt[ANIM_SCREEN_OPTION_MAGIC_LAMP2_WAVE_AMP_MIN];
    o->name = "magic_lamp2_wave_amp_min";
    o->shortDesc = N_("Magic Lamp 2 Wave Min Amplitude");
    o->longDesc =
        N_("Magic Lamp 2 waves will have random amplitudes between min and max");
    o->type = CompOptionTypeFloat;
    o->value.f = ANIM_MAGIC_LAMP2_WAVE_AMP_MIN_DEFAULT;
    o->rest.f.min = ANIM_MAGIC_LAMP2_WAVE_AMP_MIN_MIN;
    o->rest.f.max = ANIM_MAGIC_LAMP2_WAVE_AMP_MIN_MAX;
    o->rest.f.precision = ANIM_MAGIC_LAMP2_WAVE_AMP_MIN_PRECISION;

    o = &as->opt[ANIM_SCREEN_OPTION_MAGIC_LAMP2_WAVE_AMP_MAX];
    o->name = "magic_lamp2_wave_amp_max";
    o->shortDesc = N_("Magic Lamp 2 Wave Max Amplitude");
    o->longDesc =
        N_("Magic Lamp 2 waves will have random amplitudes between min and max");
    o->type = CompOptionTypeFloat;
    o->value.f = ANIM_MAGIC_LAMP2_WAVE_AMP_MAX_DEFAULT;
    o->rest.f.min = ANIM_MAGIC_LAMP2_WAVE_AMP_MAX_MIN;
    o->rest.f.max = ANIM_MAGIC_LAMP2_WAVE_AMP_MAX_MAX;
    o->rest.f.precision = ANIM_MAGIC_LAMP2_WAVE_AMP_MAX_PRECISION;

    o = &as->opt[ANIM_SCREEN_OPTION_FOCUS_WAVE_WIDTH];
    o->name = "focus_wave_width";
    o->shortDesc = N_("Focus Wave Width");
    o->longDesc = N_("Focus wave width relative to window height");
    o->type = CompOptionTypeFloat;
    o->value.f = ANIM_FOCUS_WAVE_WIDTH_DEFAULT;
    o->rest.f.min = ANIM_FOCUS_WAVE_WIDTH_MIN;
    o->rest.f.max = ANIM_FOCUS_WAVE_WIDTH_MAX;
    o->rest.f.precision = ANIM_FOCUS_WAVE_WIDTH_PRECISION;

    o = &as->opt[ANIM_SCREEN_OPTION_FOCUS_WAVE_AMP];
    o->name = "focus_wave_amp";
    o->shortDesc = N_("Focus Wave Amplitude");
    o->longDesc = N_("Focus wave amplitude relative to window height");
    o->type = CompOptionTypeFloat;
    o->value.f = ANIM_FOCUS_WAVE_AMP_DEFAULT;
    o->rest.f.min = ANIM_FOCUS_WAVE_AMP_MIN;
    o->rest.f.max = ANIM_FOCUS_WAVE_AMP_MAX;
    o->rest.f.precision = ANIM_FOCUS_WAVE_AMP_PRECISION;

    // Minimize

    o = &as->opt[ANIM_SCREEN_OPTION_MINIMIZE_EFFECT];
    o->name = "minimize_effect";
    o->shortDesc = N_("Minimize Effect");
    o->longDesc = N_("Minimize Window Effect");
    o->type = CompOptionTypeString;
    o->value.s = strdup (allEffectName[ANIM_MINIMIZE_DEFAULT]);
    o->rest.s.string = minimizeEffectName;
    o->rest.s.nString = NUM_MINIMIZE_EFFECT;

    o = &as->opt[ANIM_SCREEN_OPTION_MINIMIZE_WINDOW_TYPE];
    o->name = "minimize_window_types";
    o->shortDesc = N_("Minimize Window Types");
    o->longDesc =
        N_
        ("Window types that should animate with this effect when minimized");
    o->type = CompOptionTypeList;
    o->value.list.type = CompOptionTypeString;
    o->value.list.nValue = N_MINIMIZE_DEFAULT_WIN_TYPE;
    o->value.list.value =
        malloc (sizeof (CompOptionValue) * N_MINIMIZE_DEFAULT_WIN_TYPE);
    for (i = 0; i < N_MINIMIZE_DEFAULT_WIN_TYPE; i++)
        o->value.list.value[i].s = strdup (minimizeDefaultWinType[i]);
    o->rest.s.string = windowTypeString;
    o->rest.s.nString = nWindowTypeString;

    as->minimizeWMask = compWindowTypeMaskFromStringList (&o->value);

    o = &as->opt[ANIM_SCREEN_OPTION_MINIMIZE_DURATION];
    o->name = "minimize_duration";
    o->shortDesc = N_("Minimize Animation Duration");
    o->longDesc = N_("Minimize animation duration in seconds");
    o->type = CompOptionTypeFloat;
    o->value.f = ANIM_MINIMIZE_DURATION_DEFAULT;
    o->rest.f.min = ANIM_MINIMIZE_DURATION_MIN;
    o->rest.f.max = ANIM_MINIMIZE_DURATION_MAX;
    o->rest.f.precision = ANIM_MINIMIZE_DURATION_PRECISION;

    // Unminimize

    o = &as->opt[ANIM_SCREEN_OPTION_UNMINIMIZE_EFFECT];
    o->name = "unminimize_effect";
    o->shortDesc = N_("Unminimize Effect");
    o->longDesc = N_("Unminimize Window Effect");
    o->type = CompOptionTypeString;
    o->value.s = strdup (allEffectName[ANIM_UNMINIMIZE_DEFAULT]);
    o->rest.s.string = minimizeEffectName;
    o->rest.s.nString = NUM_MINIMIZE_EFFECT;

    o = &as->opt[ANIM_SCREEN_OPTION_UNMINIMIZE_WINDOW_TYPE];
    o->name = "unminimize_window_types";
    o->shortDesc = N_("Unminimize Window Types");
    o->longDesc =
        N_
        ("Window types that should animate with this effect when unminimized");
    o->type = CompOptionTypeList;
    o->value.list.type = CompOptionTypeString;
    o->value.list.nValue = N_MINIMIZE_DEFAULT_WIN_TYPE;
    o->value.list.value =
        malloc (sizeof (CompOptionValue) * N_MINIMIZE_DEFAULT_WIN_TYPE);
    for (i = 0; i < N_MINIMIZE_DEFAULT_WIN_TYPE; i++)
        o->value.list.value[i].s = strdup (minimizeDefaultWinType[i]);
    o->rest.s.string = windowTypeString;
    o->rest.s.nString = nWindowTypeString;

    as->unminimizeWMask = compWindowTypeMaskFromStringList (&o->value);

    o = &as->opt[ANIM_SCREEN_OPTION_UNMINIMIZE_DURATION];
    o->name = "unminimize_duration";
    o->shortDesc = N_("Unminimize Animation Duration");
    o->longDesc = N_("Unminimize animation duration in seconds");
    o->type = CompOptionTypeFloat;
    o->value.f = ANIM_UNMINIMIZE_DURATION_DEFAULT;
    o->rest.f.min = ANIM_UNMINIMIZE_DURATION_MIN;
    o->rest.f.max = ANIM_UNMINIMIZE_DURATION_MAX;
    o->rest.f.precision = ANIM_UNMINIMIZE_DURATION_PRECISION;

    // Close 1

    o = &as->opt[ANIM_SCREEN_OPTION_CLOSE1_EFFECT];
    o->name = "close1_effect";
    o->shortDesc = N_("Close Effect 1");
    o->longDesc = N_("Close Window Effect 1");
    o->type = CompOptionTypeString;
    o->value.s = strdup (allEffectName[ANIM_CLOSE1_DEFAULT]);
    o->rest.s.string = closeEffectName;
    o->rest.s.nString = NUM_CLOSE_EFFECT;

    o = &as->opt[ANIM_SCREEN_OPTION_CLOSE1_WINDOW_TYPE];
    o->name = "close1_window_types";
    o->shortDesc = N_("Close Window Types 1");
    o->longDesc =
        N_("Window types that should animate with this effect when closed");
    o->type = CompOptionTypeList;
    o->value.list.type = CompOptionTypeString;
    o->value.list.nValue = N_CLOSE1_DEFAULT_WIN_TYPE;
    o->value.list.value =
        malloc (sizeof (CompOptionValue) * N_CLOSE1_DEFAULT_WIN_TYPE);
    for (i = 0; i < N_CLOSE1_DEFAULT_WIN_TYPE; i++)
        o->value.list.value[i].s = strdup (close1DefaultWinType[i]);
    o->rest.s.string = windowTypeString;
    o->rest.s.nString = nWindowTypeString;

    as->close1WMask = compWindowTypeMaskFromStringList (&o->value);

    o = &as->opt[ANIM_SCREEN_OPTION_CLOSE1_DURATION];
    o->name = "close1_duration";
    o->shortDesc = N_("Close Animation Duration 1");
    o->longDesc = N_("Animation duration in seconds for close effect 1");
    o->type = CompOptionTypeFloat;
    o->value.f = ANIM_CLOSE1_DURATION_DEFAULT;
    o->rest.f.min = ANIM_CLOSE1_DURATION_MIN;
    o->rest.f.max = ANIM_CLOSE1_DURATION_MAX;
    o->rest.f.precision = ANIM_CLOSE1_DURATION_PRECISION;

    // Close 2

    o = &as->opt[ANIM_SCREEN_OPTION_CLOSE2_EFFECT];
    o->name = "close2_effect";
    o->shortDesc = N_("Close Effect 2");
    o->longDesc = N_("Close Window Effect 2");
    o->type = CompOptionTypeString;
    o->value.s = strdup (allEffectName[ANIM_CLOSE2_DEFAULT]);
    o->rest.s.string = closeEffectName;
    o->rest.s.nString = NUM_CLOSE_EFFECT;

    o = &as->opt[ANIM_SCREEN_OPTION_CLOSE2_WINDOW_TYPE];
    o->name = "close2_window_types";
    o->shortDesc = N_("Close Window Types 2");
    o->longDesc =
        N_("Window types that should animate with this effect when closed");
    o->type = CompOptionTypeList;
    o->value.list.type = CompOptionTypeString;
    o->value.list.nValue = N_CLOSE2_DEFAULT_WIN_TYPE;
    o->value.list.value =
        malloc (sizeof (CompOptionValue) * N_CLOSE2_DEFAULT_WIN_TYPE);
    for (i = 0; i < N_CLOSE2_DEFAULT_WIN_TYPE; i++)
        o->value.list.value[i].s = strdup (close2DefaultWinType[i]);
    o->rest.s.string = windowTypeString;
    o->rest.s.nString = nWindowTypeString;

    as->close2WMask = compWindowTypeMaskFromStringList (&o->value);

    o = &as->opt[ANIM_SCREEN_OPTION_CLOSE2_DURATION];
    o->name = "close2_duration";
    o->shortDesc = N_("Close Animation Duration 2");
    o->longDesc = N_("Animation duration in seconds for close effect 2");
    o->type = CompOptionTypeFloat;
    o->value.f = ANIM_CLOSE2_DURATION_DEFAULT;
    o->rest.f.min = ANIM_CLOSE2_DURATION_MIN;
    o->rest.f.max = ANIM_CLOSE2_DURATION_MAX;
    o->rest.f.precision = ANIM_CLOSE2_DURATION_PRECISION;

    // Create 1

    o = &as->opt[ANIM_SCREEN_OPTION_CREATE1_EFFECT];
    o->name = "create1_effect";
    o->shortDesc = N_("Create Effect 1");
    o->longDesc = N_("Create Window Effect 1");
    o->type = CompOptionTypeString;
    o->value.s = strdup (allEffectName[ANIM_CREATE1_DEFAULT]);
    o->rest.s.string = closeEffectName;
    o->rest.s.nString = NUM_CLOSE_EFFECT;

    o = &as->opt[ANIM_SCREEN_OPTION_CREATE1_WINDOW_TYPE];
    o->name = "create1_window_types";
    o->shortDesc = N_("Create Window Types 1");
    o->longDesc =
        N_("Window types that should animate with this effect when created");
    o->type = CompOptionTypeList;
    o->value.list.type = CompOptionTypeString;
    o->value.list.nValue = N_CLOSE1_DEFAULT_WIN_TYPE;
    o->value.list.value =
        malloc (sizeof (CompOptionValue) * N_CLOSE1_DEFAULT_WIN_TYPE);
    for (i = 0; i < N_CLOSE1_DEFAULT_WIN_TYPE; i++)
        o->value.list.value[i].s = strdup (close1DefaultWinType[i]);
    o->rest.s.string = windowTypeString;
    o->rest.s.nString = nWindowTypeString;

    as->create1WMask = compWindowTypeMaskFromStringList (&o->value);

    o = &as->opt[ANIM_SCREEN_OPTION_CREATE1_DURATION];
    o->name = "create1_duration";
    o->shortDesc = N_("Create Animation Duration 1");
    o->longDesc = N_("Animation duration in seconds for create effect 1");
    o->type = CompOptionTypeFloat;
    o->value.f = ANIM_CREATE1_DURATION_DEFAULT;
    o->rest.f.min = ANIM_CREATE1_DURATION_MIN;
    o->rest.f.max = ANIM_CREATE1_DURATION_MAX;
    o->rest.f.precision = ANIM_CREATE1_DURATION_PRECISION;

    // Create 2

    o = &as->opt[ANIM_SCREEN_OPTION_CREATE2_EFFECT];
    o->name = "create2_effect";
    o->shortDesc = N_("Create Effect 2");
    o->longDesc = N_("Create Window Effect 2");
    o->type = CompOptionTypeString;
    o->value.s = strdup (allEffectName[ANIM_CREATE2_DEFAULT]);
    o->rest.s.string = closeEffectName;
    o->rest.s.nString = NUM_CLOSE_EFFECT;

    o = &as->opt[ANIM_SCREEN_OPTION_CREATE2_WINDOW_TYPE];
    o->name = "create2_window_types";
    o->shortDesc = N_("Create Window Types 2");
    o->longDesc =
        N_("Window types that should animate with this effect when created");
    o->type = CompOptionTypeList;
    o->value.list.type = CompOptionTypeString;
    o->value.list.nValue = N_CLOSE2_DEFAULT_WIN_TYPE;
    o->value.list.value =
        malloc (sizeof (CompOptionValue) * N_CLOSE2_DEFAULT_WIN_TYPE);
    for (i = 0; i < N_CLOSE2_DEFAULT_WIN_TYPE; i++)
        o->value.list.value[i].s = strdup (close2DefaultWinType[i]);
    o->rest.s.string = windowTypeString;
    o->rest.s.nString = nWindowTypeString;

    as->create2WMask = compWindowTypeMaskFromStringList (&o->value);

    o = &as->opt[ANIM_SCREEN_OPTION_CREATE2_DURATION];
    o->name = "create2_duration";
    o->shortDesc = N_("Create Animation Duration 2");
    o->longDesc = N_("Animation duration in seconds for create effect 2");
    o->type = CompOptionTypeFloat;
    o->value.f = ANIM_CREATE2_DURATION_DEFAULT;
    o->rest.f.min = ANIM_CREATE2_DURATION_MIN;
    o->rest.f.max = ANIM_CREATE2_DURATION_MAX;
    o->rest.f.precision = ANIM_CREATE2_DURATION_PRECISION;

    // Focus

    o = &as->opt[ANIM_SCREEN_OPTION_FOCUS_EFFECT];
    o->name = "focus_effect";
    o->shortDesc = N_("Focus Effect");
    o->longDesc = N_("Focus Window Effect");
    o->type = CompOptionTypeString;
    o->value.s = strdup (allEffectName[ANIM_FOCUS_DEFAULT]);
    o->rest.s.string = focusEffectName;
    o->rest.s.nString = NUM_FOCUS_EFFECT;

    o = &as->opt[ANIM_SCREEN_OPTION_FOCUS_WINDOW_TYPE];
    o->name = "focus_window_types";
    o->shortDesc = N_("Focus Window Types");
    o->longDesc = N_("Window types that should animate with this effect when focused");
    o->type = CompOptionTypeList;
    o->value.list.type = CompOptionTypeString;
    o->value.list.nValue = N_FOCUS_DEFAULT_WIN_TYPE;
    o->value.list.value =
        malloc (sizeof (CompOptionValue) * N_FOCUS_DEFAULT_WIN_TYPE);
    for (i = 0; i < N_FOCUS_DEFAULT_WIN_TYPE; i++)
        o->value.list.value[i].s = strdup (focusDefaultWinType[i]);
    o->rest.s.string = windowTypeString;
    o->rest.s.nString = nWindowTypeString;

    as->focusWMask = compWindowTypeMaskFromStringList (&o->value);

    o = &as->opt[ANIM_SCREEN_OPTION_FOCUS_DURATION];
    o->name = "focus_duration";
    o->shortDesc = N_("Focus Animation Duration");
    o->longDesc = N_("Focus animation duration in seconds");
    o->type = CompOptionTypeFloat;
    o->value.f = ANIM_FOCUS_DURATION_DEFAULT;
    o->rest.f.min = ANIM_FOCUS_DURATION_MIN;
    o->rest.f.max = ANIM_FOCUS_DURATION_MAX;
    o->rest.f.precision = ANIM_FOCUS_DURATION_PRECISION;

    // Shade

    o = &as->opt[ANIM_SCREEN_OPTION_SHADE_EFFECT];
    o->name = "shade_effect";
    o->shortDesc = N_("Shade Effect");
    o->longDesc = N_("Shade Window Effect");
    o->type = CompOptionTypeString;
    o->value.s = strdup (allEffectName[ANIM_SHADE_DEFAULT]);
    o->rest.s.string = shadeEffectName;
    o->rest.s.nString = NUM_SHADE_EFFECT;

    o = &as->opt[ANIM_SCREEN_OPTION_SHADE_WINDOW_TYPE];
    o->name = "shade_window_types";
    o->shortDesc = N_("Shade Window Types");
    o->longDesc = N_("Window types that should animate with this effect when shaded");
    o->type = CompOptionTypeList;
    o->value.list.type = CompOptionTypeString;
    o->value.list.nValue = N_SHADE_DEFAULT_WIN_TYPE;
    o->value.list.value =
        malloc (sizeof (CompOptionValue) * N_SHADE_DEFAULT_WIN_TYPE);
    for (i = 0; i < N_SHADE_DEFAULT_WIN_TYPE; i++)
        o->value.list.value[i].s = strdup (shadeDefaultWinType[i]);
    o->rest.s.string = windowTypeString;
    o->rest.s.nString = nWindowTypeString;

    as->shadeWMask = compWindowTypeMaskFromStringList (&o->value);

    o = &as->opt[ANIM_SCREEN_OPTION_SHADE_DURATION];
    o->name = "shade_duration";
    o->shortDesc = N_("Shade Animation Duration");
    o->longDesc = N_("Shade animation duration in seconds");
    o->type = CompOptionTypeFloat;
    o->value.f = ANIM_SHADE_DURATION_DEFAULT;
    o->rest.f.min = ANIM_SHADE_DURATION_MIN;
    o->rest.f.max = ANIM_SHADE_DURATION_MAX;
    o->rest.f.precision = ANIM_SHADE_DURATION_PRECISION;

    // Unshade

    o = &as->opt[ANIM_SCREEN_OPTION_UNSHADE_EFFECT];
    o->name = "unshade_effect";
    o->shortDesc = N_("Unshade Effect");
    o->longDesc = N_("Unshade Window Effect");
    o->type = CompOptionTypeString;
    o->value.s = strdup (allEffectName[ANIM_UNSHADE_DEFAULT]);
    o->rest.s.string = shadeEffectName;
    o->rest.s.nString = NUM_SHADE_EFFECT;

    o = &as->opt[ANIM_SCREEN_OPTION_UNSHADE_WINDOW_TYPE];
    o->name = "unshade_window_types";
    o->shortDesc = N_("Unshade Window Types");
    o->longDesc = N_("Window types that should animate with this effect when unshaded");
    o->type = CompOptionTypeList;
    o->value.list.type = CompOptionTypeString;
    o->value.list.nValue = N_SHADE_DEFAULT_WIN_TYPE;
    o->value.list.value =
        malloc (sizeof (CompOptionValue) * N_SHADE_DEFAULT_WIN_TYPE);
    for (i = 0; i < N_SHADE_DEFAULT_WIN_TYPE; i++)
        o->value.list.value[i].s = strdup (shadeDefaultWinType[i]);
    o->rest.s.string = windowTypeString;
    o->rest.s.nString = nWindowTypeString;

    as->unshadeWMask = compWindowTypeMaskFromStringList (&o->value);

    o = &as->opt[ANIM_SCREEN_OPTION_UNSHADE_DURATION];
    o->name = "unshade_duration";
    o->shortDesc = N_("Unshade Animation Duration");
    o->longDesc = N_("Unshade animation duration in seconds");
    o->type = CompOptionTypeFloat;
    o->value.f = ANIM_UNSHADE_DURATION_DEFAULT;
    o->rest.f.min = ANIM_UNSHADE_DURATION_MIN;
    o->rest.f.max = ANIM_UNSHADE_DURATION_MAX;
    o->rest.f.precision = ANIM_UNSHADE_DURATION_PRECISION;
}

static CompOption *
animGetScreenOptions (CompScreen * screen, int *count)
{
    if (screen)
    {
        ANIM_SCREEN (screen);
    
        *count = NUM_OPTIONS (as);
        return as->opt;
    }
    else
    {
        AnimScreen * as = malloc(sizeof(AnimScreen));
        animScreenInitOptions(as);
        *count = NUM_OPTIONS(as);
        return as->opt;
    }
}

static void
objectInit (Object * object,
            float positionX, float positionY,
            float gridPositionX, float gridPositionY)
{
    object->gridPosition.x = gridPositionX;
    object->gridPosition.y = gridPositionY;

    object->position.x = positionX;
    object->position.y = positionY;

    object->offsetTexCoordForQuadBefore.x = 0;
    object->offsetTexCoordForQuadBefore.y = 0;
    object->offsetTexCoordForQuadAfter.x = 0;
    object->offsetTexCoordForQuadAfter.y = 0;
}

static void
modelInitObjects (Model * model, int x, int y, int width, int height)
{
    int gridX, gridY;
    int nGridCellsX, nGridCellsY;
    float x0, y0;

    x0 = model->scaleOrigin.x;
    y0 = model->scaleOrigin.y;

    // number of grid cells in x direction
    nGridCellsX = model->gridWidth - 1;

    if (model->forWindowEvent == WindowEventShade ||
        model->forWindowEvent == WindowEventUnshade)
    {
        // number of grid cells in y direction
        nGridCellsY = model->gridHeight - 3; // One allocated for top, one for bottom

        float winContentsHeight = height - model->topHeight - model->bottomHeight;

        //Top
        float objectY = y + (0 - y0) * model->scale.y + y0;
        
        for (gridX = 0; gridX < model->gridWidth; gridX++)
        {
            objectInit (&model->objects[gridX],
                        x + ((gridX * width / nGridCellsX) - x0) *
                        model->scale.x + x0,
                        objectY,
                        (float)gridX / nGridCellsX, 0);
        }

        // Window contents
        for (gridY = 1; gridY < model->gridHeight-1; gridY++)
        {
            float inWinY = (gridY - 1) * winContentsHeight / nGridCellsY +
                               model->topHeight;
            float gridPosY = inWinY / height;
            objectY = y + (inWinY - y0) * model->scale.y + y0;

            for (gridX = 0; gridX < model->gridWidth; gridX++)
            {
                objectInit (&model->objects[gridY * model->gridWidth + gridX],
                            x + ((gridX * width / nGridCellsX) - x0) *
                            model->scale.x + x0,
                            objectY,
                            (float)gridX / nGridCellsX, gridPosY);
            }
        }

        // Bottom (gridY is model->gridHeight-1 now)
        objectY = y + (height - y0) * model->scale.y + y0;
        
        for (gridX = 0; gridX < model->gridWidth; gridX++)
        {
            objectInit (&model->objects[gridY * model->gridWidth + gridX],
                        x + ((gridX * width / nGridCellsX) - x0) *
                        model->scale.x + x0,
                        objectY,
                        (float)gridX / nGridCellsX, 1);
        }
    }
    else
    {
        // number of grid cells in y direction
        nGridCellsY = model->gridHeight - 1;

        int i = 0;
        for (gridY = 0; gridY < model->gridHeight; gridY++)
        {
            float objectY = y + ((gridY * height / nGridCellsY) - y0) *
                            model->scale.y + y0;
            for (gridX = 0; gridX < model->gridWidth; gridX++)
            {
                objectInit (&model->objects[i],
                            x + ((gridX * width / nGridCellsX) - x0) *
                            model->scale.x + x0,
                            objectY,
                            (float)gridX / nGridCellsX,
                            (float)gridY / nGridCellsY);
                i++;
            }
        }
    }
}

static void
modelMove (Model * model, float tx, float ty)
{
    int i;

    for (i = 0; i < model->numObjects; i++)
    {
        model->objects[i].position.x += tx;
        model->objects[i].position.y += ty;
    }
}

static Model *
createModel (CompWindow *w,
             WindowEvent forWindowEvent,
             AnimEffect forAnimEffect,
             int gridWidth, int gridHeight)
{
    int x = WIN_X (w);
    int y = WIN_Y (w);
    int width = WIN_W (w);
    int height = WIN_H (w);

    Model *model;

    model = malloc (sizeof (Model));
    if (!model)
    {
        fprintf (stderr, "Animation: couldn't allocate model!\n");
        return 0;
    }
    model->magicLampWaveCount = 0;
    model->magicLampWaves = NULL;

    model->gridWidth = gridWidth;
    model->gridHeight = gridHeight;
    model->numObjects = gridWidth * gridHeight;
    model->objects = malloc (sizeof (Object) * model->numObjects);
    if (!model->objects)
    {
        fprintf (stderr, "Animation: couldn't allocate objects!\n");
        free (model);
        return 0;
    }

    // For shading
    model->forWindowEvent = forWindowEvent;
    model->topHeight = w->output.top;
    model->bottomHeight = w->output.bottom;

    model->remainderSteps = 0;

    model->scale.x = 1.0f;
    model->scale.y = 1.0f;

    model->scaleOrigin.x = 0.0f;
    model->scaleOrigin.y = 0.0f;

    model->transformed = FALSE;
    modelInitObjects (model, x, y, width, height);

    modelCalcBounds (model);

    return model;
}

static Bool
animEnsureModel (CompWindow * w,
                 WindowEvent forWindowEvent,
                 AnimEffect forAnimEffect)
{
    ANIM_WINDOW (w);
    ANIM_SCREEN (w->screen);

    int gridWidth = 2;
    int gridHeight = 2;

    if (animEffectProperties[forAnimEffect].initGridFunc)
        animEffectProperties[forAnimEffect].initGridFunc(
            as, forWindowEvent, &gridWidth, &gridHeight);

    Bool isShadeUnshadeEvent =
        (forWindowEvent == WindowEventShade ||
          forWindowEvent == WindowEventUnshade);

    Bool wasShadeUnshadeEvent = aw->model &&
        (aw->model->forWindowEvent == WindowEventShade ||
          aw->model->forWindowEvent == WindowEventUnshade);

    if (!aw->model ||
        gridWidth != aw->model->gridWidth ||
        gridHeight != aw->model->gridHeight  ||
        (isShadeUnshadeEvent != wasShadeUnshadeEvent))
    {
        if (aw->model)
        {
            if (aw->model->magicLampWaves)
                free (aw->model->magicLampWaves);
            free (aw->model->objects);
            free (aw->model);
        }
        aw->model = createModel (w,
                                 forWindowEvent, forAnimEffect,
                                 gridWidth, gridHeight);
        if (!aw->model)
            return FALSE;
    }

    if (aw->model)
        aw->model->remainderSteps = 0;

    return TRUE;
}

static void
postAnimationCleanup (CompWindow * w, Bool resetAnimation)
{
    ANIM_WINDOW (w);
    ANIM_SCREEN (w->screen);

    if (resetAnimation)
    {
        aw->curWindowEvent = WindowEventNone;
        aw->curAnimEffect = AnimEffectNone;
        aw->animOverrideProgressDir = 0;

        if (as->ppDisabling == PostprocessDisablingWindow)
            w->disablePostprocessFX = FALSE;
        if (!(as->animInProgress) &&
            as->ppDisabling == PostprocessDisablingScreen)
            w->screen->disablePostprocessFX = FALSE;
        if (aw->model)
        {
            if (aw->model->magicLampWaves)
                free (aw->model->magicLampWaves);
            aw->model->magicLampWaves = 0;
        }
    }

    aw->state = aw->newState;

    //if (aw->unmapCnt || aw->destroyCnt)
    //    releaseWindow (w);
    while (aw->unmapCnt)
    {
        unmapWindow (w);
        aw->unmapCnt--;
    }
    while (aw->destroyCnt)
    {
        destroyWindow (w);
        aw->destroyCnt--;
    }
}

static void
animPreparePaintScreen (CompScreen * s, int msSinceLastPaint)
{
    CompWindow *w;

    ANIM_SCREEN (s);

    if (as->animInProgress)
    {
        AnimWindow *aw;
        REGION region;
        Point topLeft, bottomRight;

        if (as->ppDisabling == PostprocessDisablingScreen)
            s->disablePostprocessFX = TRUE;
        else
            s->disablePostprocessFX = FALSE;

        region.rects = &region.extents;
        region.numRects = region.size = 1;

        as->animInProgress = FALSE;
        for (w = s->windows; w; w = w->next)
        {
            aw = GET_ANIM_WINDOW (w, as);

            if (aw->animRemainingTime > 0)
            {
                if (as->ppDisabling == PostprocessDisablingWindow)
                    w->disablePostprocessFX = TRUE;
                else
                    w->disablePostprocessFX = FALSE;

                if (aw->model)
                {
                    topLeft = aw->model->topLeft;
                    bottomRight = aw->model->bottomRight;

                    // If just starting, call fx init func.
                    if (aw->animRemainingTime == aw->animTotalTime &&
                        animEffectProperties[aw->curAnimEffect].initFunc)
                        animEffectProperties[aw->curAnimEffect].initFunc(s, w);

                    // Call fx step func.
                    if (animEffectProperties[aw->curAnimEffect].modelStepFunc)
                    {
                        animEffectProperties[aw->curAnimEffect].
                            modelStepFunc (s, w, msSinceLastPaint);
                    }
                    if (aw->animRemainingTime <= 0)
                    {
                        // Animation done
                        postAnimationCleanup (w, TRUE);
                    }

                    if (!(s->damageMask & COMP_SCREEN_DAMAGE_ALL_MASK))
                    {
                        if (aw->animRemainingTime > 0)
                        {
                            if (aw->model->topLeft.x < topLeft.x)
                                topLeft.x = aw->model->topLeft.x;
                            if (aw->model->topLeft.y < topLeft.y)
                                topLeft.y = aw->model->topLeft.y;
                            if (aw->model->bottomRight.x > bottomRight.x)
                                bottomRight.x = aw->model->bottomRight.x;
                            if (aw->model->bottomRight.y > bottomRight.y)
                                bottomRight.y = aw->model->bottomRight.y;
                        }
                        else
                            addWindowDamage (w);

                        region.extents.x1 = topLeft.x;
                        region.extents.y1 = topLeft.y;
                        region.extents.x2 = bottomRight.x + 0.5f;
                        region.extents.y2 = bottomRight.y + 0.5f;

                        damageScreenRegion (s, &region);
                    }
                }
                as->animInProgress |= (aw->animRemainingTime > 0);
            }

            if (aw->animRemainingTime <= 0)
            {
                if (aw->curAnimEffect!=AnimEffectNone ||
                        aw->unmapCnt > 0 ||
                        aw->destroyCnt > 0)
                    postAnimationCleanup(w,TRUE);
                aw->curWindowEvent = WindowEventNone;
                aw->curAnimEffect = AnimEffectNone;
            }
        }
    }

    UNWRAP (as, s, preparePaintScreen);
    (*s->preparePaintScreen) (s, msSinceLastPaint);
    WRAP (as, s, preparePaintScreen, animPreparePaintScreen);
}

static void
animDonePaintScreen (CompScreen * s)
{
    ANIM_SCREEN (s);

    if (as->animInProgress)
        damageScreen (s);

    UNWRAP (as, s, donePaintScreen);
    (*s->donePaintScreen) (s);
    WRAP (as, s, donePaintScreen, animDonePaintScreen);
}

static void
animAddWindowGeometry (CompWindow * w,
                       CompMatrix * matrix,
                       int nMatrix, Region region, Region clip)
{
    ANIM_WINDOW (w);
    ANIM_SCREEN (w->screen);

    // if model is lost during animation (e.g. when plugin just reloaded)
    if (aw->animRemainingTime > 0 && !aw->model)
    {
        aw->animRemainingTime = 0;
        postAnimationCleanup (w, TRUE);
    }
    // if window is being animated
    if (aw->animRemainingTime > 0 && aw->model)
    {
        BoxPtr pClip;
        int nClip;
        int nVertices, nIndices;
        GLushort *i;
        GLfloat *v;
        int x1, y1, x2, y2;
        float width, height;
        float winContentsY, winContentsHeight;
        float deformedX, deformedY;
        int nVertX, nVertY, wx, wy;
        int vSize, it;
        float gridW, gridH, x, y;
        Bool rect = TRUE;
        Bool useTextureQ = FALSE;
        Model *model = aw->model;

        // Use Q texture coordinate to avoid jagged-looking quads
        // http://www.r3.nu/~cass/qcoord/
        if (aw->curAnimEffect != AnimEffectZoom &&
            aw->curAnimEffect != AnimEffectSidekick)
            useTextureQ = TRUE;

        nClip = region->numRects;
        pClip = region->rects;

        if (nClip == 0) // nothing to do
            return;

        for (it = 0; it < nMatrix; it++)
        {
            if (matrix[it].xy != 0.0f && matrix[it].yx != 0.0f)
            {
                rect = FALSE;
                break;
            }
        }

        // window coordinates and size
        wx = WIN_X (w);
        wy = WIN_Y (w);
        width = WIN_W (w);
        height = WIN_H (w);

        // to be used if event is shade/unshade
        winContentsY = w->attrib.y;
        winContentsHeight = w->height;

        w->texUnits = nMatrix;

        if (w->vCount == 0)
        {
            // reset
            w->indexCount = 0;
            w->texCoordSize = 4;
        }
        vSize = 2 + w->texUnits * w->texCoordSize;

        nVertices = w->vCount;
        nIndices = w->indexCount;

        v = w->vertices + (nVertices * vSize);
        i = w->indices + nIndices;

        // For each clip passed to this function
        for (; nClip--; pClip++)
        {
            x1 = pClip->x1;
            y1 = pClip->y1;
            x2 = pClip->x2;
            y2 = pClip->y2;

            // printf("  x1: %d, y1: %d, x2: %d, y2: %d\n", x1, y1, x2, y2);
            
            gridW = (float)width / (model->gridWidth - 1);

            if (aw->curWindowEvent == WindowEventShade ||
                aw->curWindowEvent == WindowEventUnshade)
            {
                if (y1 < w->attrib.y) // if at top part
                {
                    gridH = model->topHeight;
                }
                else if (y2 > w->attrib.y + w->height) // if at bottom
                {
                    gridH = model->bottomHeight;
                }
                else // in window contents (only in Y coords)
                {
                    float winContentsHeight = height - model->topHeight - model->bottomHeight;
                    gridH = winContentsHeight / (model->gridHeight - 3);
                }
            }
            else
                gridH = (float)height / (model->gridHeight - 1);

            // nVertX, nVertY: number of vertices for this clip in x and y dimensions
            // + 1 to avoid running short of vertices in some cases
            nVertX = ceil((x2 - x1) / gridW) + 1 + 1;
            nVertY = ceil((y2 - y1) / gridH) + 1 + 1;

            // Allocate 4 indices for each quad
            int newIndexSize = nIndices + ((nVertX-1) * (nVertY-1) * 4);
            if (newIndexSize > w->indexSize)
            {
                if (!moreWindowIndices (w, newIndexSize))
                    return;

                i = w->indices + nIndices;
            }

            // Assign quad vertices to indices
            int jx, jy;
            for (jy = 0; jy < nVertY - 1; jy++)
            {
                for (jx = 0; jx < nVertX - 1; jx++)
                {
                    *i++ = nVertices + nVertX * (2 * jy + 1) + jx;
                    *i++ = nVertices + nVertX * (2 * jy + 1) + jx + 1;
                    *i++ = nVertices + nVertX * 2 * jy + jx + 1;
                    *i++ = nVertices + nVertX * 2 * jy + jx;

                    nIndices += 4;
                }
            }

            // Allocate vertices
            int newVertexSize = (nVertices + nVertX * (2 * nVertY - 2)) * vSize;
            if (newVertexSize > w->vertexSize)
            {
                if (!moreWindowVertices (w, newVertexSize))
                    return;

                v = w->vertices + (nVertices * vSize);
            }

            float rowTexCoordQ = 1;
            float prevRowCellWidth = 0; // this initial value won't be used
            float rowCellWidth = 0;

            // For each vertex
            for (jy = 0, y = y1; jy < nVertY; jy++)
            {
                if (y > y2)
                    y = y2;

                // Do calculations for y here to avoid repeating
                // them unnecessarily in the x loop

                float topiyFloat;
                Bool applyOffsets = TRUE;
                if (aw->curWindowEvent == WindowEventShade ||
                    aw->curWindowEvent == WindowEventUnshade)
                {
                    if (y1 < w->attrib.y) // if at top part
                    {
                        topiyFloat = (y - WIN_Y (w)) / model->topHeight;
                        topiyFloat = MIN (topiyFloat, 0.999); // avoid 1.0
                        applyOffsets = FALSE;
                    }
                    else if (y2 > w->attrib.y + w->height) // if at bottom
                    {
                        topiyFloat = (model->gridHeight - 2) +
                                     (y - (winContentsY + winContentsHeight)) /
                                     model->bottomHeight;
                        applyOffsets = FALSE;
                    }
                    else // in window contents (only in Y coords)
                    {
                        topiyFloat = 
                            (model->gridHeight - 3) *
                            (y - winContentsY) / winContentsHeight + 1;
                    }
                }
                else
                {
                    topiyFloat =
                        (model->gridHeight - 1) * (y - wy) / height;
                }
                // topiy should be at most (model->gridHeight - 2)
                int topiy = (int)(topiyFloat + 1e-4);

                if (topiy == model->gridHeight - 1)
                    topiy--;
                int bottomiy = topiy + 1;
                float iny = topiyFloat - topiy;

                // End of calculations for y

                for (jx = 0, x = x1; jx < nVertX; jx++)
                {
                    if (x > x2)
                        x = x2;

                    // find containing grid cell (leftix rightix) x (topiy bottomiy)
                    float leftixFloat = (model->gridWidth - 1) * (x - wx) / width;
                    int leftix = (int)(leftixFloat + 1e-4);

                    if (leftix == model->gridWidth - 1)
                        leftix--;
                    int rightix = leftix + 1;

                    // Objects that are at top, bottom, left, right corners of quad
                    Object * objToTopLeft =
                        & (model->objects[topiy * model->gridWidth + leftix]);
                    Object * objToTopRight =
                        & (model->objects[topiy * model->gridWidth + rightix]);
                    Object * objToBottomLeft =
                        & (model->objects[bottomiy * model->gridWidth + leftix]);
                    Object * objToBottomRight =
                        & (model->objects[bottomiy * model->gridWidth + rightix]);

                    // find position in cell by taking remainder of flooring
                    float inx = leftixFloat - leftix;

                    // Interpolate to find deformed coordinates

                    float hor1x =
                        (1 - inx) * objToTopLeft->position.x +
                        inx       * objToTopRight->position.x;
                    float hor1y =
                        (1 - inx) * objToTopLeft->position.y +
                        inx       * objToTopRight->position.y;
                    float hor2x =
                        (1 - inx) * objToBottomLeft->position.x +
                        inx       * objToBottomRight->position.x;
                    float hor2y =
                        (1 - inx) * objToBottomLeft->position.y +
                        inx       * objToBottomRight->position.y;

                    deformedX = (1 - iny) * hor1x + iny * hor2x;
                    deformedY = (1 - iny) * hor1y + iny * hor2y;

                    if (useTextureQ)
                    {
                        if (jx == 1)
                            rowCellWidth = deformedX - v[-2];

                        if (jy > 0 && jx == 1)  // do only once per row for all rows except row 0
                        {
                            rowTexCoordQ = (rowCellWidth / prevRowCellWidth);

                            v[-3] = rowTexCoordQ;   // update first column
                            v[-6] *= rowTexCoordQ;  // (since we didn't know rowTexCoordQ before)
                            v[-5] *= rowTexCoordQ;
                        }
                    }
                    if (rect)
                    {
                        for (it = 0; it < nMatrix; it++)
                        {
                            float offsetY = 0;
                            if (applyOffsets && y < y2)
                                offsetY = objToTopLeft->offsetTexCoordForQuadAfter.y;

                            *v++ = COMP_TEX_COORD_X (&matrix[it], x);
                            *v++ = COMP_TEX_COORD_Y (&matrix[it], y + offsetY);
                            *v++ = 0;
                            if (useTextureQ)
                            {
                                *v++ = rowTexCoordQ;    // Q texture coordinate

                                if (0 < jy && jy < nVertY - 1)
                                {
                                    // copy first 3 texture coords to duplicate row
                                    memcpy (v - 4 + nVertX * vSize, v - 4,
                                            3 * sizeof (GLfloat));
                                    *(v - 1 + nVertX * vSize) = 1;  // Q texture coordinate
                                }
                                if (applyOffsets && /*0 < jy && */objToTopLeft->offsetTexCoordForQuadBefore.y != 0)
                                {
                                    // After copying to next row, update texture y coord
                                    // by following object's offset
                                    offsetY = objToTopLeft->offsetTexCoordForQuadBefore.y;
                                    v[-3] = COMP_TEX_COORD_Y (&matrix[it], y + offsetY);
                                }
                                if (jx > 0)     // since column 0 is updated when jx == 1
                                {
                                    v[-4] *= rowTexCoordQ;
                                    v[-3] *= rowTexCoordQ;
                                }
                            }
                            else
                            {
                                *v++ = 1;

                                if (0 < jy && jy < nVertY - 1)
                                {
                                    // copy first 3 texture coords to duplicate row
                                    memcpy (v - 4 + nVertX * vSize, v - 4,
                                            3 * sizeof (GLfloat));

                                    *(v - 1 + nVertX * vSize) = 1;  // Q texture coordinate
                                }
                                if (applyOffsets && objToTopLeft->offsetTexCoordForQuadBefore.y != 0)
                                {
                                    // After copying to next row, update texture y coord
                                    // by following object's offset
                                    offsetY = objToTopLeft->offsetTexCoordForQuadBefore.y;
                                    v[-3] = COMP_TEX_COORD_Y (&matrix[it], y + offsetY);
                                }
                            }
                        }
                    }
                    else
                    {
                        for (it = 0; it < nMatrix; it++)
                        {
                            float offsetY = 0;
                            if (applyOffsets && y < y2)
                                offsetY = objToTopLeft->offsetTexCoordForQuadAfter.y;

                            *v++ = COMP_TEX_COORD_XY (&matrix[it], x, y + offsetY);
                            *v++ = COMP_TEX_COORD_YX (&matrix[it], x, y + offsetY);
                            *v++ = 0;
                            if (useTextureQ)
                            {
                                *v++ = rowTexCoordQ;    // Q texture coordinate

                                if (0 < jy && jy < nVertY - 1)
                                {
                                    // copy first 3 texture coords to duplicate row
                                    memcpy (v - 4 + nVertX * vSize, v - 4,
                                            3 * sizeof (GLfloat));
                                    *(v - 1 + nVertX * vSize) = 1;  // Q texture coordinate
                                }
                                if (applyOffsets && objToTopLeft->offsetTexCoordForQuadBefore.y != 0)
                                {
                                    // After copying to next row, update texture y coord
                                    // by following object's offset
                                    offsetY = objToTopLeft->offsetTexCoordForQuadBefore.y;
                                    v[-4] = COMP_TEX_COORD_XY (&matrix[it], x, y + offsetY);
                                    v[-3] = COMP_TEX_COORD_YX (&matrix[it], x, y + offsetY);
                                }
                                if (jx > 0)     // column t should be updated when jx is t+1
                                {
                                    v[-4] *= rowTexCoordQ;
                                    v[-3] *= rowTexCoordQ;
                                }
                            }
                            else
                            {
                                *v++ = 1;

                                if (0 < jy && jy < nVertY - 1)
                                {
                                    // copy first 3 texture coords to duplicate row
                                    memcpy (v - 4 + nVertX * vSize, v - 4,
                                            3 * sizeof (GLfloat));
                                    *(v - 1 + nVertX * vSize) = 1;      // Q texture coordinate
                                }
                                if (applyOffsets && objToTopLeft->offsetTexCoordForQuadBefore.y != 0)
                                {
                                    // After copying to next row, update texture y coord
                                    // by following object's offset
                                    offsetY = objToTopLeft->offsetTexCoordForQuadBefore.y;
                                    v[-4] = COMP_TEX_COORD_XY (&matrix[it], x, y + offsetY);
                                    v[-3] = COMP_TEX_COORD_YX (&matrix[it], x, y + offsetY);
                                }
                            }
                        }
                    }
                    *v++ = deformedX;
                    *v++ = deformedY;

                    if (0 < jy && jy < nVertY - 1)
                        memcpy (v - 2 + nVertX * vSize, v - 2,
                                2 * sizeof (GLfloat));

                    nVertices++;

                    // increment x properly (so that coordinates fall on grid intersections)
                    x = rightix * gridW + wx;
                }
                if (useTextureQ)
                    prevRowCellWidth = rowCellWidth;

                if (0 < jy && jy < nVertY - 1)
                {
                    v += nVertX * vSize;    // skip the duplicate row
                    nVertices += nVertX;
                }

                // increment y properly (so that coordinates fall on grid intersections)
                if (aw->curWindowEvent == WindowEventShade ||
                    aw->curWindowEvent == WindowEventUnshade)
                {
                    y += gridH;
                }
                else
                {
                    y = bottomiy * gridH + wy;
                }
            }
        }
        w->vCount = nVertices;
        w->indexCount = nIndices;
    }
    else
    {
        UNWRAP (as, w->screen, addWindowGeometry);
        (*w->screen->addWindowGeometry) (w, matrix, nMatrix, region, clip);
        WRAP (as, w->screen, addWindowGeometry, animAddWindowGeometry);
    }
}

static void
animDrawWindowGeometry (CompWindow * w)
{
    ANIM_WINDOW (w);

    if (aw->animRemainingTime > 0)
    {
        int texUnit = w->texUnits;
        int currentTexUnit = 0;
        int stride = 2 + texUnit * w->texCoordSize;
        GLfloat *vertices = w->vertices + (stride - 2);

        stride *= sizeof (GLfloat);

        //glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

        glVertexPointer (2, GL_FLOAT, stride, vertices);

        while (texUnit--)
        {
            if (texUnit != currentTexUnit)
            {
                w->screen->clientActiveTexture (GL_TEXTURE0_ARB + texUnit);
                glEnableClientState (GL_TEXTURE_COORD_ARRAY);
                currentTexUnit = texUnit;
            }
            vertices -= w->texCoordSize;
            glTexCoordPointer (w->texCoordSize, GL_FLOAT, stride, vertices);
        }

        glDrawElements (GL_QUADS, w->indexCount, GL_UNSIGNED_SHORT,
                        w->indices);

        // disable all texture coordinate arrays except 0
        texUnit = w->texUnits;
        if (texUnit > 1)
        {
            while (--texUnit)
            {
                (*w->screen->clientActiveTexture) (GL_TEXTURE0_ARB + texUnit);
                glDisableClientState (GL_TEXTURE_COORD_ARRAY);
            }

            (*w->screen->clientActiveTexture) (GL_TEXTURE0_ARB);
        }
    }
    else
    {
        ANIM_SCREEN (w->screen);

        UNWRAP (as, w->screen, drawWindowGeometry);
        (*w->screen->drawWindowGeometry) (w);
        WRAP (as, w->screen, drawWindowGeometry, animDrawWindowGeometry);
    }
}

static Bool
animPaintWindow (CompWindow * w,
                 const WindowPaintAttrib * attrib,
                 Region region, unsigned int mask)
{
    Bool status;

    ANIM_SCREEN (w->screen);
    ANIM_WINDOW (w);

    if (aw->animRemainingTime > 0)
    {
        w->indexCount = 0;

        WindowPaintAttrib wAttrib = *attrib;

        if (mask & PAINT_WINDOW_SOLID_MASK)
            return FALSE;

        mask |= PAINT_WINDOW_TRANSFORMED_MASK;

        wAttrib.xScale = 1.0f;
        wAttrib.yScale = 1.0f;

        if (animEffectProperties[aw->curAnimEffect].updateWindowAttribFunc)
            animEffectProperties[aw->curAnimEffect].
                updateWindowAttribFunc (as, aw, &wAttrib);

        UNWRAP (as, w->screen, paintWindow);
        status = (*w->screen->paintWindow) (w, &wAttrib, region, mask);
        WRAP (as, w->screen, paintWindow, animPaintWindow);
    }
    else
    {
        UNWRAP (as, w->screen, paintWindow);
        status = (*w->screen->paintWindow) (w, attrib, region, mask);
        WRAP (as, w->screen, paintWindow, animPaintWindow);
    }

    return status;
}

static Bool
animGetWindowIconGeometry (CompWindow * w, XRectangle * rect)
{
    Atom actual;
    int result, format;
    unsigned long n, left;
    unsigned char *data;

    ANIM_DISPLAY (w->screen->display);

    result = XGetWindowProperty (w->screen->display->display, w->id,
                                 ad->winIconGeometryAtom,
                                 0L, 4L, FALSE, XA_CARDINAL, &actual, &format,
                                 &n, &left, &data);

    if (result == Success && n && data)
    {
        if (n == 4)
        {
            unsigned long *geometry = (unsigned long *)data;

            rect->x = geometry[0];
            rect->y = geometry[1];
            rect->width = geometry[2];
            rect->height = geometry[3];

            XFree (data);

            return TRUE;
        }

        XFree (data);
    }

    return FALSE;
}

static void
animHandleEvent (CompDisplay * d, XEvent * event)
{
    Window activeWindow = 0;
    CompWindow *w;

    ANIM_DISPLAY (d);

    switch (event->type)
    {
    case PropertyNotify:
        if (event->xproperty.atom == d->winActiveAtom)
            activeWindow = d->activeWindow;
        break;
    case MapNotify:
        w = findWindowAtDisplay (d, event->xmap.window);
        if (w)
        {
            ANIM_WINDOW (w);

            if (aw->animRemainingTime > 0)
            {
                aw->state = aw->newState;
            }
            while (aw->unmapCnt)
            {
                unmapWindow (w);
                aw->unmapCnt--;
            }
        }
        break;
    case DestroyNotify:
        w = findWindowAtDisplay (d, event->xunmap.window);
        if (w)
        {
            ANIM_WINDOW (w);
            aw->destroyCnt++;
            w->destroyRefCnt++;
            addWindowDamage (w);
        }
        break;
    case UnmapNotify:
        w = findWindowAtDisplay (d, event->xunmap.window);
        if (w)
        {
            ANIM_SCREEN (w->screen);

            if (w->pendingUnmaps)       // Normal -> Iconic
            {
                ANIM_WINDOW (w);
                if (w->shaded)
                {
                    // SHADE event!

                    //printf("SHADE event! %X\n", (unsigned)aw);

                    aw->nowShaded = TRUE;
 
                    if (as->shadeEffect &&
                        (as->shadeWMask & w->type))
                    {
                        Bool startingNew = TRUE;
                        if (aw->curWindowEvent != WindowEventNone)
                        {
                            if (aw->curWindowEvent != WindowEventUnshade)
                                postAnimationCleanup (w, TRUE);
                            else
                            {
                                // Play the unshade effect backwards from where it left
                                aw->animRemainingTime =
                                    aw->animTotalTime - aw->animRemainingTime;
     
                                // avoid window remains
                                if (aw->animRemainingTime <= 0)
                                    aw->animRemainingTime = 1;

                                startingNew = FALSE;
                                if (aw->animOverrideProgressDir == 0)
                                    aw->animOverrideProgressDir = 2;
                                else if (aw->animOverrideProgressDir == 1)
                                    aw->animOverrideProgressDir = 0;
                            }
                        }

                        as->animInProgress = TRUE;
                        aw->curWindowEvent = WindowEventShade;

                        if (startingNew)
                        {
                            aw->curAnimEffect = as->shadeEffect;
                            aw->animTotalTime =
                                as->opt[ANIM_SCREEN_OPTION_SHADE_DURATION].
                                value.f * 1000;
                            aw->animRemainingTime = aw->animTotalTime;
                        }
                        if (!animEnsureModel (w, WindowEventShade, aw->curAnimEffect))
                        {
                            postAnimationCleanup (w, TRUE);
                        }

                        aw->unmapCnt++;
                        w->unmapRefCnt++;

                        addWindowDamage (w);
                    }
                }
                else if (!w->invisible && as->minimizeEffect &&
                         (as->minimizeWMask & w->type))
                {
                    // MINIMIZE event!

                    //printf("MINIMIZE event! %X\n", (unsigned)aw);

                    Bool startingNew = TRUE;
                    if (aw->curWindowEvent != WindowEventNone)
                    {
                        if (aw->curWindowEvent != WindowEventUnminimize)
                            postAnimationCleanup (w, TRUE);
                        else
                        {
                            // Play the unminimize effect backwards from where it left
                            aw->animRemainingTime =
                                aw->animTotalTime - aw->animRemainingTime;
 
                            // avoid window remains
                            if (aw->animRemainingTime == 0)
                                aw->animRemainingTime = 1;

                            startingNew = FALSE;
                            if (aw->animOverrideProgressDir == 0)
                                aw->animOverrideProgressDir = 2;
                            else if (aw->animOverrideProgressDir == 1)
                                aw->animOverrideProgressDir = 0;
                        }
                    }

                    aw->newState = IconicState;
                    as->animInProgress = TRUE;
                    aw->curWindowEvent = WindowEventMinimize;

                    if (startingNew)
                    {
                        aw->curAnimEffect = as->minimizeEffect;
                        aw->animTotalTime =
                            as->opt[ANIM_SCREEN_OPTION_MINIMIZE_DURATION].
                            value.f * 1000;
                        aw->animRemainingTime = aw->animTotalTime;
                    }
                    if (!animEnsureModel (w, WindowEventMinimize, aw->curAnimEffect))
                    {
                        postAnimationCleanup (w, TRUE);
                    }
                    else
                    {
                        if (!animGetWindowIconGeometry (w, &aw->icon))
                        {
                            // minimize to bottom-center if there is no window list
                            aw->icon.x = w->screen->width / 2;
                            aw->icon.y = w->screen->height;
                            aw->icon.width = 100;
                            aw->icon.height = 20;
                        }
                        if ((aw->curAnimEffect == AnimEffectZoom ||
                             aw->curAnimEffect == AnimEffectSidekick) &&
                            as->opt[ANIM_SCREEN_OPTION_ZOOM_FROM_CENTER].value.b)
                        {
                            aw->icon.x = WIN_X (w) + WIN_W (w) / 2 - aw->icon.width / 2;
                            aw->icon.y = WIN_Y (w) + WIN_H (w) / 2 - aw->icon.height / 2;
                        }

                        aw->unmapCnt++;
                        w->unmapRefCnt++;

                        addWindowDamage (w);
                    }
                }
            }
            else                // X -> Withdrawn
            {
                ANIM_WINDOW (w);

                AnimEffect windowsCloseEffect = AnimEffectNone;
                int whichClose = 1; // either 1 or 2

                if (as->close1Effect && (as->close1WMask & w->type))
                    windowsCloseEffect = as->close1Effect;
                else if (as->close2Effect && (as->close2WMask & w->type))
                {
                    windowsCloseEffect = as->close2Effect;
                    whichClose = 2;
                }

                // CLOSE event!

                if (windowsCloseEffect)
                {
                    int tmpSteps = 0;

                    //printf("CLOSE event! %X\n", (unsigned)aw);

                    Bool startingNew = TRUE;

                    if (aw->animRemainingTime > 0 &&
                        aw->curWindowEvent != WindowEventCreate)
                    {
                        tmpSteps = aw->animRemainingTime;
                        aw->animRemainingTime = 0;
                    }
                    if (aw->curWindowEvent != WindowEventNone)
                    {
                        if (aw->curWindowEvent == WindowEventCreate)
                        {
                            // Play the create effect backward from where it left
                            aw->animRemainingTime =
                                aw->animTotalTime - aw->animRemainingTime;

                            // avoid window remains
                            if (aw->animRemainingTime <= 0)
                                aw->animRemainingTime = 1;

                            startingNew = FALSE;
                            if (aw->animOverrideProgressDir == 0)
                                aw->animOverrideProgressDir = 2;
                            else if (aw->animOverrideProgressDir == 1)
                                aw->animOverrideProgressDir = 0;
                        }
                        else if (aw->curWindowEvent == WindowEventClose)
                        {
                            if (aw->animOverrideProgressDir == 2)
                            {
                                aw->animRemainingTime = tmpSteps;
                                startingNew = FALSE;
                            }
                        }
                        else
                        {
                            postAnimationCleanup (w, TRUE);
                        }
                    }

                    aw->state = NormalState;
                    aw->newState = WithdrawnState;
                    as->animInProgress = TRUE;
                    aw->curWindowEvent = WindowEventClose;

                    if (startingNew)
                    {
                        aw->curAnimEffect = windowsCloseEffect;
                        aw->animTotalTime =
                            as->opt[whichClose == 1 ?
                            ANIM_SCREEN_OPTION_CLOSE1_DURATION :
                            ANIM_SCREEN_OPTION_CLOSE2_DURATION].value.f * 1000;
                        aw->animRemainingTime = aw->animTotalTime;
                    }
                    if (!animEnsureModel (w, WindowEventClose, aw->curAnimEffect))
                    {
                        postAnimationCleanup (w, TRUE);
                    }
                    else if (getMousePointerXY
                             (w->screen, &aw->icon.x, &aw->icon.y))
                    {
                        aw->icon.width = FAKE_ICON_SIZE;
                        aw->icon.height = FAKE_ICON_SIZE;
                        aw->icon.x -= FAKE_ICON_SIZE / 2;
                        aw->icon.y -= FAKE_ICON_SIZE / 2;
                        if ((aw->curAnimEffect == AnimEffectZoom ||
                             aw->curAnimEffect == AnimEffectSidekick) &&
                            as->opt[ANIM_SCREEN_OPTION_ZOOM_FROM_CENTER].value.b)
                        {
                            aw->icon.x = WIN_X (w) + WIN_W (w) / 2 - aw->icon.width / 2;
                            aw->icon.y = WIN_Y (w) + WIN_H (w) / 2 - aw->icon.height / 2;
                        }

                        aw->unmapCnt++;
                        w->unmapRefCnt++;

                        addWindowDamage (w);
                    }
                }
                else if ((as->create1Effect && (as->create1WMask & w->type)) ||
                         (as->create2Effect && (as->create2WMask & w->type)))
                {
                    // stop the current animation and prevent it from rewinding

                    if (aw->animRemainingTime > 0 &&
                        aw->curWindowEvent != WindowEventCreate)
                    {
                        aw->animRemainingTime = 0;
                    }
                    if ((aw->curWindowEvent != WindowEventNone) &&
                        (aw->curWindowEvent != WindowEventClose))
                    {
                        postAnimationCleanup (w, TRUE);
                    }

                    // set some properties to make sure this window will use the
                    // correct create effect the next time it's "created"

                    aw->state = NormalState;
                    aw->newState = WithdrawnState;
                    as->animInProgress = TRUE;
                    aw->curWindowEvent = WindowEventClose;

                    aw->unmapCnt++;
                    w->unmapRefCnt++;

                    addWindowDamage (w);
                }
            }
        }
        break;
    default:
        break;
    }

    UNWRAP (ad, d, handleEvent);
    (*d->handleEvent) (d, event);
    WRAP (ad, d, handleEvent, animHandleEvent);

    switch (event->type)
    {
    case PropertyNotify:
        if (event->xproperty.atom == d->winActiveAtom)
        {
            if (d->activeWindow != activeWindow)
            {
                w = findWindowAtDisplay (d, d->activeWindow);
                if (w)
                {
                    ANIM_WINDOW (w);
                    ANIM_SCREEN (w->screen);

                    if (aw->curWindowEvent != WindowEventNone)
                        return;

                    if ((as->focusWMask & w->type) && as->focusEffect &&
                        // On unminimization, focus event is fired first.
                        // When this happens and minimize is in progress,
                        // don't prevent rewinding of minimize when unminimize is fired
                        // right after this focus event.
                        aw->curWindowEvent != WindowEventMinimize &&
                        animEnsureModel (w, WindowEventFocus, as->focusEffect))
                    {
                        // FOCUS event!

                        //printf("FOCUS event! %X\n", (unsigned)aw);

                        if (aw->curWindowEvent != WindowEventNone)
                        {
                            postAnimationCleanup (w, TRUE);
                        }

                        as->animInProgress = TRUE;
                        aw->curWindowEvent = WindowEventFocus;
                        aw->curAnimEffect = as->focusEffect;
                        aw->animTotalTime =
                            as->opt[ANIM_SCREEN_OPTION_FOCUS_DURATION].value.
                            f * 1000;
                        aw->animRemainingTime = aw->animTotalTime;

                        damageScreen (w->screen);
                    }
                }
            }
        }

    default:
        break;
    }
}

static Bool
animDamageWindowRect (CompWindow * w, Bool initial, BoxPtr rect)
{
    Bool status;

    ANIM_SCREEN (w->screen);

    if (initial)                // Unminimize or Create
    {
        ANIM_WINDOW (w);

        if (aw->state == IconicState)
        {
            if (!w->invisible && as->unminimizeEffect && (as->unminimizeWMask & w->type))
            {
                // UNMINIMIZE event!

                //printf("UNMINIMIZE event! %X\n", (unsigned)aw);

                Bool startingNew = TRUE;
                if (aw->curWindowEvent != WindowEventNone)
                {
                    if (aw->curWindowEvent != WindowEventMinimize)
                        postAnimationCleanup (w, TRUE);
                    else
                    {
                        // Play the minimize effect backwards from where it left
                        aw->animRemainingTime =
                            aw->animTotalTime - aw->animRemainingTime;

                        // avoid window remains
                        if (aw->animRemainingTime <= 0)
                            aw->animRemainingTime = 1;

                        startingNew = FALSE;
                        if (aw->animOverrideProgressDir == 0)
                            aw->animOverrideProgressDir = 1;
                        else if (aw->animOverrideProgressDir == 2)
                            aw->animOverrideProgressDir = 0;
                    }
                }
                as->animInProgress = TRUE;
                aw->curWindowEvent = WindowEventUnminimize;

                if (startingNew)
                {
                    aw->curAnimEffect = as->unminimizeEffect;
                    aw->animTotalTime =
                        as->opt[ANIM_SCREEN_OPTION_UNMINIMIZE_DURATION].value.
                        f * 1000;
                    aw->animRemainingTime = aw->animTotalTime;
                }

                if (animEnsureModel (w, WindowEventUnminimize, aw->curAnimEffect))
                {
                    if (!animGetWindowIconGeometry (w, &aw->icon))
                    {
                        // minimize to bottom-center if there is no window list
                        aw->icon.x = w->screen->width / 2;
                        aw->icon.y = w->screen->height;
                        aw->icon.width = 100;
                        aw->icon.height = 20;
                    }
                    if ((aw->curAnimEffect == AnimEffectZoom ||
                         aw->curAnimEffect == AnimEffectSidekick) &&
                        as->opt[ANIM_SCREEN_OPTION_ZOOM_FROM_CENTER].value.b)
                    {
                        aw->icon.x = WIN_X (w) + WIN_W (w) / 2 - aw->icon.width / 2;
                        aw->icon.y = WIN_Y (w) + WIN_H (w) / 2 - aw->icon.height / 2;
                    }
                    addWindowDamage (w);
                }
                else
                    postAnimationCleanup (w, TRUE);
            }
        }
        else if (aw->nowShaded)
        {
            // UNSHADE event!
            //printf("UNSHADE event! %X\n", (unsigned)aw);

            aw->nowShaded = FALSE;

            if (as->unshadeEffect &&
                (as->unshadeWMask & w->type))
            {
                Bool startingNew = TRUE;
                if (aw->curWindowEvent != WindowEventNone)
                {
                    if (aw->curWindowEvent != WindowEventShade)
                        postAnimationCleanup (w, TRUE);
                    else
                    {
                        // Play the shade effect backwards from where it left
                        aw->animRemainingTime =
                            aw->animTotalTime - aw->animRemainingTime;

                        // avoid window remains
                        if (aw->animRemainingTime <= 0)
                            aw->animRemainingTime = 1;

                        startingNew = FALSE;
                        if (aw->animOverrideProgressDir == 0)
                            aw->animOverrideProgressDir = 1;
                        else if (aw->animOverrideProgressDir == 2)
                            aw->animOverrideProgressDir = 0;
                    }
                }
                as->animInProgress = TRUE;
                aw->curWindowEvent = WindowEventUnshade;

                if (startingNew)
                {
                    aw->curAnimEffect = as->unshadeEffect;
                    aw->animTotalTime =
                        as->opt[ANIM_SCREEN_OPTION_UNSHADE_DURATION].value.
                        f * 1000;
                    aw->animRemainingTime = aw->animTotalTime;
                }

                if (animEnsureModel (w, WindowEventUnshade, aw->curAnimEffect))
                    addWindowDamage (w);
                else
                    postAnimationCleanup (w, TRUE);
            }
        }
        else if (aw->state != NormalState && !w->invisible)
        {
            AnimEffect windowsCreateEffect = AnimEffectNone;

            int whichCreate = 1; // either 1 or 2
            
            if (as->create1Effect && (as->create1WMask & w->type))
                windowsCreateEffect = as->create1Effect;
            else if (as->create2Effect && (as->create2WMask & w->type))
            {
                windowsCreateEffect = as->create2Effect;
                whichCreate = 2;
            }

            if (windowsCreateEffect &&
                getMousePointerXY (w->screen, &aw->icon.x, &aw->icon.y))
            {
                // CREATE event!

                //printf("CREATE event! %X\n", (unsigned)aw);

                Bool startingNew = TRUE;
                if (aw->curWindowEvent != WindowEventNone)
                {
                    if (aw->curWindowEvent != WindowEventClose)
                        postAnimationCleanup (w, TRUE);
                    else
                    {
                        // Play the close effect backwards from where it left
                        aw->animRemainingTime =
                            aw->animTotalTime - aw->animRemainingTime;

                        // avoid window remains
                        if (aw->animRemainingTime == 0)
                            aw->animRemainingTime = 1;

                        startingNew = FALSE;
                        if (aw->animOverrideProgressDir == 0)
                            aw->animOverrideProgressDir = 1;
                        else if (aw->animOverrideProgressDir == 2)
                            aw->animOverrideProgressDir = 0;
                    }
                }
                as->animInProgress = TRUE;
                aw->curWindowEvent = WindowEventCreate;

                if (startingNew)
                {
                    aw->curAnimEffect = windowsCreateEffect;
                    aw->animTotalTime =
                        as->opt[whichCreate == 1 ?
                            ANIM_SCREEN_OPTION_CREATE1_DURATION :
                            ANIM_SCREEN_OPTION_CREATE2_DURATION].value.f *
                        1000;
                    aw->animRemainingTime = aw->animTotalTime;
                }

                aw->icon.width = FAKE_ICON_SIZE;
                aw->icon.height = FAKE_ICON_SIZE;
                aw->icon.x -= FAKE_ICON_SIZE / 2;
                aw->icon.y -= FAKE_ICON_SIZE / 2;
                if ((aw->curAnimEffect == AnimEffectZoom ||
                     aw->curAnimEffect == AnimEffectSidekick) &&
                    as->opt[ANIM_SCREEN_OPTION_ZOOM_FROM_CENTER].value.b)
                {
                    aw->icon.x = WIN_X (w) + WIN_W (w) / 2 - aw->icon.width / 2;
                    aw->icon.y = WIN_Y (w) + WIN_H (w) / 2 - aw->icon.height / 2;
                }
                aw->state = IconicState;        // we're doing this as a hack, it may not be necessary

                if (animEnsureModel (w, WindowEventCreate, aw->curAnimEffect))
                    addWindowDamage (w);
                else
                    postAnimationCleanup (w, TRUE);
            }
        }

        aw->newState = NormalState;
    }

    UNWRAP (as, w->screen, damageWindowRect);
    status = (*w->screen->damageWindowRect) (w, initial, rect);
    WRAP (as, w->screen, damageWindowRect, animDamageWindowRect);

    return status;
}

static void
animWindowResizeNotify (CompWindow * w)
{
    ANIM_SCREEN (w->screen);
    ANIM_WINDOW (w);

    if (aw->animRemainingTime > 0)
    {
        aw->animRemainingTime = 0;
        postAnimationCleanup (w, TRUE);
    }

    if (aw->model)
    {
        modelInitObjects (aw->model,
                          WIN_X (w), WIN_Y (w), WIN_W (w), WIN_H (w));
    }

    aw->state = w->state;

    UNWRAP (as, w->screen, windowResizeNotify);
    (*w->screen->windowResizeNotify) (w);
    WRAP (as, w->screen, windowResizeNotify, animWindowResizeNotify);
}

static void
animWindowMoveNotify (CompWindow * w, int dx, int dy, Bool immediate)
{
    ANIM_SCREEN (w->screen);
    ANIM_WINDOW (w);

    CompWindow *w2;

    if (aw->animRemainingTime > 0 && aw->grabbed)
    {
        aw->animRemainingTime = 0;
        if (as->animInProgress)
        {
            as->animInProgress = FALSE;
            for (w2 = w->screen->windows; w2; w2 = w2->next)
            {
                AnimWindow *aw2;
                aw2 = GET_ANIM_WINDOW (w2, as);
                if (aw2->animRemainingTime > 0)
                {
                    as->animInProgress = TRUE;
                    break;
                }
            }
        }
        postAnimationCleanup (w, TRUE);
    }
    else if (aw->model)
    {
        modelMove (aw->model, dx, dy);
    }

    UNWRAP (as, w->screen, windowMoveNotify);
    (*w->screen->windowMoveNotify) (w, dx, dy, immediate);
    WRAP (as, w->screen, windowMoveNotify, animWindowMoveNotify);
}

static void
animWindowGrabNotify (CompWindow * w,
                      int x, int y, unsigned int state, unsigned int mask)
{
    ANIM_SCREEN (w->screen);
    ANIM_WINDOW (w);

    aw->grabbed = TRUE;

    UNWRAP (as, w->screen, windowGrabNotify);
    (*w->screen->windowGrabNotify) (w, x, y, state, mask);
    WRAP (as, w->screen, windowGrabNotify, animWindowGrabNotify);
}

static void
animWindowUngrabNotify (CompWindow *w)
{
    ANIM_SCREEN (w->screen);
    ANIM_WINDOW (w);

    aw->grabbed = FALSE;

    UNWRAP (as, w->screen, windowUngrabNotify);
    (*w->screen->windowUngrabNotify) (w);
    WRAP (as, w->screen, windowUngrabNotify, animWindowUngrabNotify);
}

static void
animSetWindowScale (CompWindow *w,
                      float      xScale,
                      float      yScale)
{
    ANIM_WINDOW (w);
    ANIM_SCREEN (w->screen);

    UNWRAP (as, w->screen, setWindowScale);
    (*w->screen->setWindowScale) (w, xScale, yScale);
    WRAP (as, w->screen, setWindowScale, animSetWindowScale);

    if (animEnsureModel (w, WindowEventFocus, as->focusEffect))
        // since focus is the most likely effect next
    {
        if (aw->model->scale.x != xScale ||
            aw->model->scale.y != yScale)
        {
            aw->model->scale.x = xScale;
            aw->model->scale.y = yScale;

            aw->model->scaleOrigin.x = w->output.left;
            aw->model->scaleOrigin.y = w->output.top;

            modelInitObjects (aw->model,
                              WIN_X (w), WIN_Y (w),
                              WIN_W (w), WIN_H (w));
        }

        if (aw->model->scale.x != 1.0f || aw->model->scale.y != 1.0f)
            aw->model->transformed = 1;
        else
            aw->model->transformed = 0;
    }
}

static Bool
animPaintScreen (CompScreen * s,
                 const ScreenPaintAttrib * sAttrib,
                 Region region, int output, unsigned int mask)
{
    Bool status;

    ANIM_SCREEN (s);

    if (as->animInProgress)
        mask |= PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK;

    UNWRAP (as, s, paintScreen);
    status = (*s->paintScreen) (s, sAttrib, region, output, mask);
    WRAP (as, s, paintScreen, animPaintScreen);

    return status;
}

static Bool
animInitDisplay (CompPlugin * p, CompDisplay * d)
{
    AnimDisplay *ad;

    ad = malloc (sizeof (AnimDisplay));
    if (!ad)
        return FALSE;

    ad->screenPrivateIndex = allocateScreenPrivateIndex (d);
    if (ad->screenPrivateIndex < 0)
    {
        free (ad);
        return FALSE;
    }

    ad->wmHintsAtom = XInternAtom (d->display, "WM_HINTS", FALSE);
    ad->winIconGeometryAtom =
        XInternAtom (d->display, "_NET_WM_ICON_GEOMETRY", 0);

    WRAP (ad, d, handleEvent, animHandleEvent);

    d->privates[displayPrivateIndex].ptr = ad;

    return TRUE;
}

static void
animFiniDisplay (CompPlugin * p, CompDisplay * d)
{
    ANIM_DISPLAY (d);

    freeScreenPrivateIndex (d, ad->screenPrivateIndex);

    UNWRAP (ad, d, handleEvent);

    free (ad);
}

static Bool
animInitScreen (CompPlugin * p, CompScreen * s)
{
    AnimScreen *as;

    ANIM_DISPLAY (s->display);

    as = calloc (1,sizeof (AnimScreen));
    if (!as)
        return FALSE;

    as->windowPrivateIndex = allocateWindowPrivateIndex (s);
    if (as->windowPrivateIndex < 0)
    {
        free (as);
        return FALSE;
    }
    as->animInProgress = FALSE;
    as->minimizeEffect = ANIM_MINIMIZE_DEFAULT;
    as->unminimizeEffect = ANIM_UNMINIMIZE_DEFAULT;
    as->create1Effect = ANIM_CREATE1_DEFAULT;
    as->create2Effect = ANIM_CREATE2_DEFAULT;
    as->close1Effect = ANIM_CLOSE1_DEFAULT;
    as->close2Effect = ANIM_CLOSE2_DEFAULT;
    as->focusEffect = ANIM_FOCUS_DEFAULT;
    as->shadeEffect = ANIM_SHADE_DEFAULT;
    as->unshadeEffect = ANIM_UNSHADE_DEFAULT;

    as->ppDisabling = ANIM_DISABLE_PP_FX_DEFAULT;

    animScreenInitOptions (as);

    WRAP (as, s, preparePaintScreen, animPreparePaintScreen);
    WRAP (as, s, donePaintScreen, animDonePaintScreen);
    WRAP (as, s, paintScreen, animPaintScreen);
    WRAP (as, s, paintWindow, animPaintWindow);
    WRAP (as, s, damageWindowRect, animDamageWindowRect);
    WRAP (as, s, addWindowGeometry, animAddWindowGeometry);
    WRAP (as, s, drawWindowGeometry, animDrawWindowGeometry);
    WRAP (as, s, windowResizeNotify, animWindowResizeNotify);
    WRAP (as, s, windowMoveNotify, animWindowMoveNotify);
    WRAP (as, s, windowGrabNotify, animWindowGrabNotify);
    WRAP (as, s, windowUngrabNotify, animWindowUngrabNotify);
    WRAP (as, s, setWindowScale, animSetWindowScale);

    s->privates[ad->screenPrivateIndex].ptr = as;

    return TRUE;
}

static void
animFiniScreen (CompPlugin * p, CompScreen * s)
{
    ANIM_SCREEN (s);

    freeWindowPrivateIndex (s, as->windowPrivateIndex);

    free (as->opt[ANIM_SCREEN_OPTION_MINIMIZE_EFFECT].value.s);
    free (as->opt[ANIM_SCREEN_OPTION_UNMINIMIZE_EFFECT].value.s);
    free (as->opt[ANIM_SCREEN_OPTION_CREATE1_EFFECT].value.s);
    free (as->opt[ANIM_SCREEN_OPTION_CREATE2_EFFECT].value.s);
    free (as->opt[ANIM_SCREEN_OPTION_CLOSE1_EFFECT].value.s);
    free (as->opt[ANIM_SCREEN_OPTION_CLOSE2_EFFECT].value.s);
    free (as->opt[ANIM_SCREEN_OPTION_FOCUS_EFFECT].value.s);
    free (as->opt[ANIM_SCREEN_OPTION_SHADE_EFFECT].value.s);
    free (as->opt[ANIM_SCREEN_OPTION_UNSHADE_EFFECT].value.s);

    UNWRAP (as, s, preparePaintScreen);
    UNWRAP (as, s, donePaintScreen);
    UNWRAP (as, s, paintScreen);
    UNWRAP (as, s, paintWindow);
    UNWRAP (as, s, damageWindowRect);
    UNWRAP (as, s, addWindowGeometry);
    UNWRAP (as, s, drawWindowGeometry);
    UNWRAP (as, s, windowResizeNotify);
    UNWRAP (as, s, windowMoveNotify);
    UNWRAP (as, s, windowGrabNotify);
    UNWRAP (as, s, windowUngrabNotify);
    UNWRAP (as, s, setWindowScale);

    free (as);
}

static Bool
animInitWindow (CompPlugin * p, CompWindow * w)
{
    AnimWindow *aw;

    ANIM_SCREEN (w->screen);

    aw = calloc (1,sizeof (AnimWindow));
    if (!aw)
        return FALSE;

    aw->model = 0;
    aw->state = w->state;
    aw->animRemainingTime = 0;
    aw->curAnimEffect = AnimEffectNone;
    aw->curWindowEvent = WindowEventNone;
    aw->animOverrideProgressDir = 0;
    w->indexCount = 0;

    aw->unmapCnt = 0;
    aw->destroyCnt = 0;

    aw->grabbed = FALSE;

    if (w->shaded)
    {
        aw->state = aw->newState = NormalState;
        aw->nowShaded = TRUE;
    }
    else
    {
        aw->state = aw->newState = animGetWindowState (w);
        aw->nowShaded = FALSE;
    }

    w->privates[as->windowPrivateIndex].ptr = aw;

    return TRUE;
}

static void
animFiniWindow (CompPlugin * p, CompWindow * w)
{
    ANIM_WINDOW (w);

    postAnimationCleanup(w,FALSE);
    
    if (aw->model)
    {
        if (aw->model->magicLampWaves)
            free (aw->model->magicLampWaves);
        free (aw->model->objects);
        free (aw->model);
    }

    while (aw->unmapCnt--)
        unmapWindow (w);

    free (aw);
}

static Bool
animInit (CompPlugin * p)
{
    displayPrivateIndex = allocateDisplayPrivateIndex ();
    if (displayPrivateIndex < 0)
        return FALSE;

    return TRUE;
}

static void
animFini (CompPlugin * p)
{
    if (displayPrivateIndex >= 0)
        freeDisplayPrivateIndex (displayPrivateIndex);
}

CompPluginDep animDeps[] = {
    {CompPluginRuleBefore, "fade"}
    ,
    {CompPluginRuleBefore, "cube"}
    ,
    {CompPluginRuleBefore, "scale"}
};


static int
animGetVersion (CompPlugin *plugin,
		int	   version)
{
    return ABIVERSION;
}



CompPluginVTable animVTable = {
    "animation",
    N_("Animations"),
    N_("Use various animations as window effects"),
    animGetVersion,
    animInit,
    animFini,
    animInitDisplay,
    animFiniDisplay,
    animInitScreen,
    animFiniScreen,
    animInitWindow,
    animFiniWindow,
    0,
    0,
    animGetScreenOptions,
    animSetScreenOption,
    animDeps,
    sizeof (animDeps) / sizeof (animDeps[0])
};

CompPluginVTable *
getCompPluginInfo (void)
{
    return &animVTable;
}
